"""
Policy managers
---------------

All email policy managers inherit from :class:`~.EmailPolicy`, which provides a
fair amount of base functionality useful to its subclasses.  So far, all but
the :class:`~.SPFEnforcementPolicy` are contained here.  That one has
extra dependencies which are thus kept isolated.  Find it in
:mod:`.spf_policy`.

"""
import time
from contextlib import contextmanager
from collections import deque
from typing import List, Dict, Union, Optional, Tuple
import functools
import redis
import logging
from expiring_dict import ExpiringDict
from chapps.config import CHAPPSConfig, env
from chapps.signals import NullSenderException, NoRecipientsException
from chapps.models import Quota, SDAStatus, PolicyResponse
from chapps.util import PostfixPolicyRequest
from chapps.outbound import OutboundPPR
from chapps.inbound import InboundPPR

if env.get("CHAPPS_DB_MODULE", None) == "mysql":
    from chapps.adapter import (
        MariaDBQuotaAdapter as OBQAdapter,  # outbound quota
        MariaDBSenderDomainAuthAdapter as SDAAdapter,
        MariaDBInboundFlagsAdapter as IBFAdapter,
    )
else:
    from chapps.sqla_adapter import (
        SQLAQuotaAdapter as OBQAdapter,
        SQLASenderDomainAuthAdapter as SDAAdapter,
        SQLAInboundFlagsAdapter as IBFAdapter,
    )

policy_response = PolicyResponse.policy_response  # a parameterized decorator
logger = logging.getLogger(__name__)
seconds_per_day = 3600 * 24
SENTINEL_TIMEOUT = 0.1
TIME_FORMAT = "%d %b %Y %H:%M:%S %z"


# There are a number of commented debug statements in this module
# This is for convenience, because in production these routines need
# to be as performant as possible, but these messages are often very
# helpful for diagnosing problems during testing and debugging


class EmailPolicy:
    """Abstract policy manager

    Subclasses must:
      * set class attribute `redis_key_prefix` to a unique value.
      * define an instance method called :meth:`~.approve_policy_request`.

    This abstract superclass provides a standard framework for constructing
    Redis keys, since the main purpose of the policy manager is to make
    decisions about email by consulting Redis.

    Instance attributes:

      :config: :class:`chapps.config.CHAPPSConfig` either passed in
        during initialization or inherited from the environment

      :params: :class:`chapps.util.AttrDict` corresponding to the config
        for the policy manager

      :sentinel: a :class:`redis.Sentinel` handle, or `None` if using
        only Redis

      :redis: a :class:`redis.Redis` handle

    """

    redis_key_prefix = "chapps"
    """A placeholder value, since this class is abstract

    Subclasses should set this to a unique prefix identifying the policy
    manager.  For examples, see the included subclasses.

    """

    @staticmethod
    def rediskey(prefix: str, *args):
        """Format a string to serve as a Redis key for arbitrary data

        :param str prefix: a prefix unique to the policy manager (subclass)
        :param List[str] args: a list of strings to use to construct the rest
          of the key

        In CHAPPS, each policy has its own prefix.  What other data the policy
        uses to construct the key is not relevant to any other entities, though
        it must be sent as a string.

        This routine simply joins up all the tokens with colon (`:`)
        characters, so it is not recommended to use them as part of the
        key-components (although it should 'just work').

        """
        return f"{ prefix }:{ ':'.join( args ) }"

    @classmethod
    def _fmtkey(cls, *args):
        """Convenience classmethod for Redis key construction

        :param List[str] args: a list of key components

        Subclasses may use this method which automatically discovers the
        prefix.

        :meta public:
        """
        return cls.rediskey(cls.redis_key_prefix, *args)

    def __init__(self, cfg: CHAPPSConfig = None):
        """Sets up a new policy manager

        :param chapps.config.CHAPPSConfig cfg: optional :class:`CHAPPSConfig`
          object for config override

        Store the config and get the params for the specific policy class,
        which are in a config block named for the class.  Using that config, set
        the policy manager up with a Redis handle, and an instance cache (from
        :class:`expiring_dict.ExpiringDict`)

        """
        self.config = cfg or CHAPPSConfig.get_config()
        self.params = self.config.get_block(self.__class__.__name__)
        self._adapter = None
        self.sentinel = None
        self.redis = self._redis()  # pass True to get read-only
        self.instance_cache = ExpiringDict(3)  # entries expire after 3 seconds

    @contextmanager
    def _adapter_handle(self):
        """Context manager for obtaining a database handle

        In order to acquire policy configuration data, the policy manager must
        be able to reach the RDBMS or other policy-config data store.  One of
        the policy manager's priciple jobs is to obtain this data from the
        database and stuff it into Redis for future reference.

        Adapter configuration, in terms of how to access the database, is
        obtained from the config object.

        .. todo::

          It is clear now that the adapter classes should also accept an
          optional config argument, and then use it for default values, so that
          this routine need not enumerate all the options.

        :meta public:
        """
        if not self._adapter:
            self._adapter = self.adapter_class(cfg=self.config)
        try:
            yield self._adapter
        finally:
            if getattr(self._adapter, "conn", None):
                self._adapter.conn.close()
                self._adapter = None

    @contextmanager
    def _control_data_storage_context(self):
        """Context manager for storing control data in Redis

        This is the most basic of routines, provided for policies which store
        only one item in Redis, perhaps an option flag for instance.

        The context manager yields a callable which expects up to three
        arguments:

        .. code::

            dsc(token, setting, expire=seconds_per_day)

        :token: is a token unique to the resource along with the setting.  The
                `redis_key_prefix` is automatically prepended to the token.  If
                the token is otherwise compound, it should be compounded
                beforehand and presented as a string.

        :setting: is the value to be stored in Redis

        :expire: the Redis entry's TTL in seconds; defaults to 24hr

        This pattern is used throughout the policy managers to handle making
        Redis settings in pipelines.  Some policy managers declare overrides
        for this method, in order to automate creation of compound keys or to
        store more than one piece of data at once.  Where a policy has a
        per-domain enforcement flag, that flag is generally being stored on a
        Redis key formed by tacking the `redis_key_prefix` onto the front of
        the domain name.

        In practice, to use this context manager, capture its yielded output
        and call it in order to store data:

        .. code::

            resource_settings_map = {d: True for d in domains}
            with policy._control_data_storage_context() as dsc:
                for domain, option in resource_settings_map.items():
                    dsc(domain, option)

        If there are a number of settings to create, make sure to place the
        loop within the context so that all the settings will be submitted as
        part of the same pipeline.

        """
        pipe = self.redis.pipeline()
        fmtkey = self._fmtkey

        def _dsc(token, setting, expire=seconds_per_day):
            pipe.set(fmtkey(token), setting, ex=expire)

        try:
            yield _dsc
        finally:
            pipe.execute()
            pipe.reset()

    def _redis(self, read_only: bool = False):
        """Get a Redis handle, possibly from Sentinel

        :param bool read_only: if Sentinel is in use, get a read-only handle

        If you're not using Sentinel, the `read_only` parameter is
        meaningless.

        """
        try:
            if self.config.redis.sentinel_servers and not self.sentinel:
                self.sentinel = redis.Sentinel(
                    [
                        s.split(":")
                        for s in self.config.redis.sentinel_servers.split(" ")
                    ],
                    socket_timeout=SENTINEL_TIMEOUT,
                )
            if self.sentinel:
                if read_only:
                    rh = self.sentinel.slave_for(
                        self.config.redis.sentinel_dataset,
                        socket_timeout=SENTINEL_TIMEOUT,
                    )
                else:
                    rh = self.sentinel.master_for(
                        self.config.redis.sentinel_dataset,
                        socket_timeout=SENTINEL_TIMEOUT,
                    )
                return rh
        except AttributeError:
            pass
        return redis.Redis(
            host=self.config.redis.server, port=self.config.redis.port
        )

    def approve_policy_request(
        self, ppr: PostfixPolicyRequest, **opts
    ) -> Union[str, bool]:
        """Determine a policy outcome based on the PPR provided

        This routine may return a boolean PASS/FAIL response, or it may for
        some policy classes return a string, which represents the policy
        outcome and is suitable for sending to Postfix.

        The result of the policy approval is cached based on the instance value
        provided by Postfix.  The memoization is done here in the superclass in
        order to avoid duplication of memoization code.

        """
        response = self.instance_cache.get(ppr.instance, None)
        if response is None:
            response = self._approve_policy_request(ppr, **opts)
            self.instance_cache[ppr.instance] = response
        return response

    def _approve_policy_request(
        self, ppr: PostfixPolicyRequest, **opts
    ) -> Union[str, bool]:
        """Placeholder method which must be implemented by subclasses."""
        raise NotImplementedError(
            "Subclasses of EmailPolicy must implement this function."
        )


class PostfixActions:
    """Superclass for Postfix action adapters"""

    @staticmethod
    def dunno(*args, **kwargs):
        """Return the Postfix directive `DUNNO`"""
        return "DUNNO"

    @staticmethod
    def okay(*args, **kwargs):
        """Return the Postfix directive `OK`"""
        return "OK"

    ok = okay
    """`ok()` is an alias for `okay()`"""

    @staticmethod
    def defer_if_permit(msg, *args, **kwargs):
        """
        Return the Postfix `DEFER_IF_PERMIT` directive with the provided
        message
        """
        return f"DEFER_IF_PERMIT {msg}"

    @staticmethod
    def reject(msg, *args, **kwargs):
        """
        Return the Postfix `REJECT` directive along with the provided message
        """
        return f"REJECT {msg}"

    @staticmethod
    def prepend(*args, **kwargs):
        """
        Return the Postfix `PREPEND` directive.
        Include the header to prepend as keyword-argument `prepend`
        """
        new_header = kwargs.get("prepend", None)
        if new_header is None or len(new_header) < 5:
            raise ValueError(
                f"Prepended header expected to be at least 5 chars in length."
            )
        return f"PREPEND {new_header}"

    def __init__(self, cfg=None):
        """
        Optionally supply a :py:class:`chapps.config.CHAPPSConfig` instance as
        the first argument.
        """
        self.config = cfg or CHAPPSConfig.get_config()
        self.params = self.config  # later this is overridden, in subclasses

    def _get_closure_for(
        self, decision: str, *, passing: Optional[bool] = None
    ):
        """Setup the prescribed closure for generating SMTP action directives"""
        action_config = getattr(self.params, decision, None)
        if not action_config:
            raise ValueError(
                f"Action config for {self.__class__.__name__} does not contain a key named {decision}"
            )
        action_tokens = action_config.split(" ")
        action = action_tokens[0]
        try:
            i = int(action)  #  if the first token is a number, its a directive
        except ValueError:  #  first token was a string, and therefore refers to a method
            # look for predefined or memoized version
            af = getattr(self, action, None)
            if af:
                return af
            # no local version found, find function reference
            action_func = getattr(PostfixActions, action, None)
            if not action_func:
                action_func = getattr(self.__class__, action, None)
            if not action_func:
                raise NotImplementedError(
                    f"Action {action} is not implemented by PostfixActions"
                    f" or by {self.__class__.__name__}"
                )
        else:
            # construct closure from configured message string
            action_func = (
                lambda reason, ppr, *args, **kwargs: action_config.format(
                    reason=reason
                )
            )
        passing = (
            (action_func in [self.reject, self.defer_if_permit])
            if passing is None
            else passing
        )
        action_func = policy_response(passing, action)(action_func)
        # memoize the action function for quicker reference next time
        setattr(self, action, action_func)
        return action_func

    def _get_message_for(self, decision, config_name=None):
        """Grab a status message for a decision from the config, optionally with another name"""
        msg_key = config_name or decision
        msg = getattr(self, msg_key, None)
        if not msg:
            raise ValueError(
                f"There is no key {msg_key} in the config for {self.__class__.__name__} or its policy"
            )
        return msg

    def _mangle_action(self, action):
        """
        Policy decisions which are also reserved words need to be altered.
        Currently, this routine handles only the action 'pass'
        """
        if action == "pass":
            return "passing"
        return action

    def action_for(self, *args, **kwargs):
        """Abstract method which must be implemented in subclasses.

        This method is intended to map responses from a policy
        manager onto Postfix directives.

        Not all policy managers return only yes/no answers.  Some, like
        :py:class:`chapps.spf_policy.SPFEnforcementPolicy`, return a handful of
        different possible values, and so there must be a mechanism for
        allowing sites to determine what happens when each of those different
        outcomes occurs.

        """
        raise NotImplementedError(
            f"Subclasses of {self.__class__.__name__} must define the method"
            " action_for() for themselves, to map policy module responses"
            " (decisions) onto Postfix action directives."
        )


class PostfixPassfailActions(PostfixActions):
    """Postfix Actions adapter for PASS/FAIL policy responses.

    Many policies return `True` if the email should be accepted/forwarded and
    return `False` if the email should be rejected/dropped.  This class
    encapsulates the common case, and includes some logic to extract precise
    instructions from the config.

    """

    def __init__(self, cfg=None):
        super().__init__(cfg)

    def _get_closure_for(
        self, decision, *, passing: bool = None, msg_key: str = None
    ):
        """Create a closure for formatting these messages and store it on
        self.<decision>, and also return it
        """
        msg_key = msg_key or decision
        msg = getattr(self.params, msg_key, None)
        if not msg:
            raise ValueError(
                f"The key '{msg_key}' is not defined in the config for"
                f" {self.__class__.__name__} or its policy"
            )
        msg_tokens = msg.split(" ")
        msg_text = ""
        if msg_tokens[0] == "OK":
            func = PostfixActions.okay
        elif msg_tokens[0] == "DUNNO":
            func = PostfixActions.dunno
        elif msg_tokens[0] == "DEFER_IF_PERMIT":
            func = PostfixActions.defer_if_permit
            msg_text = " ".join(msg_tokens[1:])
        elif msg_tokens[0] == "REJECT" or msg_tokens[0] == "554":
            func = PostfixActions.reject
            msg_text = " ".join(msg_tokens[1:])
        else:
            raise NotImplementedError(
                "Pass-fail closure creation for Postfix directive"
                f" {msg_tokens[0]} is not yet available."
            )
        action = policy_response(
            (
                passing
                if passing is not None
                else (func != PostfixActions.reject)
            ),
            decision,
        )(self.__prepend_action_with_message(func, msg_text))
        setattr(self, decision, action)
        return action

    def __prepend_action_with_message(self, func, prepend_msg_text):
        """Wrap an action func in order to prepend an additional message"""
        # avoiding use of nonlocal required if definition is embedded inline
        # in calling procedure
        def action(message="", *args, **kwargs):
            msg_text = prepend_msg_text
            if len(message) > 0:
                msg_text = " ".join([msg_text, message])
            return func(msg_text, *args, **kwargs)

        return action

    def action_for(self, pf_result):
        """Return an action closure for a pass/fail policy

        Evaluates its argument `pf_result` as a boolean and returns the
        action closure for 'passing' if True, otherwise the action closure for
        'fail'.  To provide backwards-compatibility with older versions, and to
        allow for more descriptive configuration elements, the actions may be
        attached to keys named `acceptance_message` or `rejection_message`
        instead of `passing` and `fail` respectively.  This is only true
        of policies with action factories inheriting from
        :py:class:`.PostfixPassfailActions`

        """
        if pf_result:  # True / pass
            action_name = "passing"
        else:  # False / fail
            action_name = "fail"
        return getattr(self, action_name, None)

    def __getattr__(self, attrname, *args, **kwargs):
        """Allow use of old config elements with long descriptive names."""
        attrname = self._mangle_action(attrname)
        if attrname == "passing":
            msg_key = "acceptance_message"
        elif attrname == "fail":
            msg_key = "rejection_message"
        else:
            raise NotImplementedError(
                f"Pass-fail actions do not include {attrname}"
            )
        return self._get_closure_for(
            attrname, msg_key=msg_key, passing=(attrname == "passing")
        )


class PostfixOQPActions(PostfixPassfailActions):
    """Postfix Action translator for :py:class:`chapps.policy.OutboundQuotaPolicy`"""

    def __init__(self, cfg=None):
        """
        Optionally provide an instance of :py::class`chapps.config.CHAPPSConfig`.

        All this class does is wire up `self.config` to
        point at the :py:class:`chapps.policy.OutboundQuotaPolicy` config block.
        """
        super().__init__(cfg)
        self.params = self.config.policy_oqp


class PostfixGRLActions(PostfixPassfailActions):
    """Postfix Action translator for :py:class:`chapps.policy.GreylistingPolicy`"""

    def __init__(self, cfg=None):
        """
        Optionally provide an instance of :py:class:`chapps.config.CHAPPSConfig`.

        All this class does is wire up `self.config` to
        point at the :py:class:`chapps.policy.GreylistingPolicy` config block.
        """
        super().__init__(cfg)
        self.params = self.config.policy_grl


class InboundPolicy(EmailPolicy):
    adapter_class = IBFAdapter  # inbound flags adapter indirection

    def domain_option_key(self, ppr: InboundPPR):
        """Return the Redis key for the domain's Greylisting option

        Uses the first of the list of tokenized recipients.  Generally,
        inbound mail is expected to contain only one recipient per email.
        """
        return self._domain_option_key(ppr.recipient_domain)

    def _domain_option_key(self, recipient_domain):
        return self._fmtkey(recipient_domain)

    def _store_control_data(self, domain: str, flag: bool):
        with self._control_data_storage_context() as dsc:
            dsc(domain, 1 if flag else 0)

    def _whitelisted(self, ppr: InboundPPR) -> bool:
        return ppr.helo_match(self.config.helo_whitelist)


class GreylistingPolicy(InboundPolicy):
    """Policy manager which implements greylisting

    `Greylisting <https://en.wikipedia.org/wiki/Greylisting_(email)>`_ is a
    `well-defined <https://datatracker.ietf.org/doc/html/rfc6647>`_ and
    frequently-implemented pattern.  This implementation stores the tracking
    information in Redis.

    Instance attributes (in addition to those of :class:`.EmailPolicy`):

      :min_defer: minimum time between retries, in seconds

      :cache_ttl: how long to store tracking data, in seconds

      :allow_after: success threshold, after which the client may be
        whitelisted

    """

    redis_key_prefix = "grl"
    """Greylisting Redis key prefix"""

    def __init__(
        self,
        cfg: CHAPPSConfig = None,
        *,
        minimum_deferral: int = 60,
        cache_ttl: int = seconds_per_day,
        auto_allow_after: int = None,
    ):
        """Initialize a greylisting policy manager

        :param chapps.config.CHAPPSConfig cfg: optional config override

        :param int minimum_deferral: min time between retries, in seconds

        :param int cache_ttl: tracking data cache expiration time in seconds

        :param int auto_allow_after: number of successful attempts needed
          before source client is considered trusted, and no longer incurs
          deferrals

        """
        super().__init__(cfg)
        self.actions = PostfixGRLActions(self.config)
        self.min_defer = minimum_deferral
        self.cache_ttl = cache_ttl
        self.allow_after = (
            auto_allow_after
            if auto_allow_after is not None
            else self.params.whitelist_threshold
        )
        if self.cache_ttl <= self.min_defer:
            logger.warning(
                f"Cache TTL (={datetime.timedelta(seconds=self.cache_ttl)}) is not allowed to be smaller than or equal to the minimum deferral window (={datetime.timedelta(seconds=self.min_defer)}).  Defaulting to 24 hr."
            )
            self.cache_ttl = seconds_per_day
        if self.min_defer > 60 * 15:
            logger.warning(
                f"It may be unreasonable to expect the sending server to defer for more than 15 minutes. (={self.min_defer/60.0:.2f}m)"
            )
        if self.allow_after == 0:
            logger.warning(f"Sender auto-approval is turned off.")
        elif self.allow_after < 2:
            logger.warning(
                f"Sender auto-approval is set to a fairly low threshold. (={self.allow_after})"
            )

    def tuple_key(self, ppr: PostfixPolicyRequest) -> str:
        """Return the greylisting tuple as a Redis key

        The names of the values taken from `ppr` are as follows (in order):

             - `client_address`
             - `sender`
             - `recipient`

        """
        return self._tuple_key(ppr.client_address, ppr.sender, ppr.recipient)

    def _tuple_key(self, client_address, sender, recipient):
        return self._fmtkey(client_address, sender, recipient)

    def client_key(self, ppr: PostfixPolicyRequest):
        """Return the greylisting client key

        This key indicates whether the client has enough successful
        resubmissions to be whitelisted.

        """
        return self._client_key(ppr.client_address)

    def _client_key(self, client_address):
        return self._fmtkey(client_address)

    def acquire_policy_for(self, ppr: InboundPPR):
        with self._adapter_handle() as adapter:
            result = adapter.do_greylisting_on(ppr.recipient_domain)
        logger.debug(
            "Got greylisting option flag "
            + str(result)
            + " from RDBMS for domain "
            + ppr.recipient_domain
        )
        self._store_control_data(ppr.recipient_domain, 1 if result else 0)
        return result

    def _approve_policy_request(self, ppr: InboundPPR, **opts):
        """Perform greylisting

        .. todo::

            It would be possible to allow domains to set the whitelisting
            threshhold, since this routine has to obtain option flag data from
            the policy config source at this point anyway.  Because we're also
            concerned with time between attempts, we could also do some
            time-based things here, such as starting to refuse clients who
            retry much too quickly, implementing per-domain rules about how
            frequently a client is allowed to send email to their addresses,
            etc.

        :meta public:
        """
        if self._whitelisted(ppr):  # forward without SPF header
            logger.info(f"Whitelisting traffic from {ppr.helo_name}.")
            return "DUNNO"  # needs to be the string
        option_set, tuple_seen, client_tally = None, None, None
        try:
            option_set, tuple_seen, client_tally = self._get_control_data(ppr)
            if option_set is None:
                option_set = self.acquire_policy_for(ppr)
        except NoRecipientsException:
            logger.exception(f"No recipient in PPR {ppr.instance}.")
            return self.actions.action_for(False)("", ppr=ppr)
        except Exception:  # pragma: no cover
            logger.exception("UNEXPECTED")
            logger.debug(
                f"Returning denial for {ppr.instance} (unexpected exception)."
            )
        if opts.get("force", False):
            option_set = True
        if not option_set:
            logger.debug(
                "Not enforcing greylisting for domain "
                f"{ppr.recipient_domain or 'N/A'}"
            )
            return "DUNNO"  # needs to be a string
        # if not whitelisting, client_tally will be None
        if client_tally is not None and client_tally >= self.allow_after:
            self._update_client_tally(ppr)
            return self.actions.action_for(True)("", ppr=ppr)
        if tuple_seen:
            # The tuple is recognized; need to check if it was long enough ago
            now = time.time()
            if now - tuple_seen >= self.min_defer:
                # the email will be approved; some housekeeping is necessary
                self._update_client_tally(ppr)
                return self.actions.action_for(True)("", ppr=ppr)
        # if we get here, the tuple either isn't stored or was stored too
        # recently; either way, we update it
        self._update_tuple(ppr)
        return self.actions.action_for(False)("", ppr=ppr)

    def _get_control_data(self, ppr: InboundPPR):
        """Extract data from Redis in order to answer the policy request"""
        now = time.time()
        tuple_key = self.tuple_key(ppr)
        client_key = self.client_key(ppr)
        option_key = self.domain_option_key(ppr)
        logger.debug(
            f"Redis keys: tuple={tuple_key} opt={option_key} client={client_key} (zrange)"
        )
        pipe = self.redis.pipeline()
        pipe.zremrangebyscore(client_key, 0, now - float(self.cache_ttl))
        pipe.get(tuple_key)
        pipe.get(option_key)
        if self.allow_after > 0:
            pipe.zrange(client_key, 0, -1)

        result = pipe.execute()
        logger.debug(f"Redis result: {result!r}")
        tuple_bits = result[1]
        option_bits = result[2]
        if len(result) == 4:
            client_tally_bits = result[3]

        tuple_seen = (
            float(tuple_bits) if tuple_bits else None
        )  # UNIX epoch time
        option_set = int(option_bits) if option_bits else None
        client_tally = None
        if self.allow_after > 0 and client_tally_bits:
            client_tally = len(client_tally_bits)
        return (option_set, tuple_seen, client_tally)

    def _update_client_tally(self, ppr: InboundPPR):
        """Update client reliability score in Redis

        When an email is allowed, increment the reliability score of the
        client.

        """
        if self.allow_after == 0:  # if we're not keeping a tally, return
            return
        now = time.time()
        client_key = self.client_key(ppr)
        with self.redis.pipeline() as pipe:
            pipe.zadd(client_key, {ppr.instance: now})
            pipe.zremrangebyrank(
                client_key, 0, -(self.allow_after + 2)
            )  # keep one extra
            pipe.expire(client_key, self.cache_ttl)
            pipe.execute()

    def _update_tuple(self, ppr: InboundPPR):
        """Set or update a greylisting tuple in Redis"""
        self.redis.setex(self.tuple_key(ppr), self.cache_ttl, time.time())


class OutboundQuotaPolicy(EmailPolicy):
    """Policy manager which implements an outbound quota limitation

    Outbound email is controlled based on the count of (attempted)
    transmissions in the last 24 hours.  Some parameters are provided to
    fine-tune the behavior of the limiting algorithm.

    Instance attributes (in addition to those of :class:`.EmailPolicy`):

      :interval: number of seconds to store transmission attemps, and
          to use for quota evaluation; defaults to one day

      :counting_recipients: a boolean determined from the config; whether to
        count each recipient of a multi-recipient email as a separate
        transmission for quota purposes

      :min_delta: defaults to 0; if set, the number of seconds which
        must elapse between send attempts.  **Currently experimental**

    """

    adapter_class = OBQAdapter
    redis_key_prefix = "oqp"
    """OutboundQuotaPolicy Redis prefix"""

    def __init__(self, cfg=None, *, enforcement_interval=None, min_delta=0):
        """Set up an outbound quota policy manager

        :param chapps.config.CHAPPSConfig cfg: optional config override

        :param int enforcement_interval: number of seconds to store
          transmission attemps, and to use for quota evaluation; defaults to
          one day

        :param int min_delta: Minimum time which must pass between
          transmission attempts; defaults to 5 seconds to prevent spamming.
          Set to 0 to disable

        """
        super().__init__(cfg)  # sets attrs 'config', 'params', and 'redis'
        self.interval = (
            enforcement_interval if enforcement_interval else seconds_per_day
        )
        if hasattr(self.params, "min_delta"):
            self.min_delta = self.params.min_delta
        elif min_delta:
            self.min_delta = min_delta
        else:
            self.min_delta = 0
        self.min_delta = float(self.min_delta)
        self.counting_recipients = (
            self.params.counting_recipients
            if hasattr(self.params, "counting_recipients")
            else False
        )

    @contextmanager
    def _control_data_storage_context(self):
        """Atomic context manager for Redis updates

        Yields a closure which takes the tuple (email, quota, margin) and
        adds it to a Redis pipeline, which will set it in Redis once the
        context is closed.

        Intended to be used as a context manager, like so:

        .. code::python
            with self._control_data_storage_context() as store:
                store(user_identifier, quota_count, margin_count)

        This is most useful when collections of data are involved, but also
        encapsulates and hides the pipeline management foo required by the
        Redis library.

        """
        pipe = self.redis.pipeline()
        fmtkey = self._fmtkey

        def _dsc(user, quota, margin):
            pipe.set(fmtkey(user, "limit"), quota, ex=seconds_per_day)
            pipe.set(fmtkey(user, "margin"), margin, ex=seconds_per_day)

        try:
            yield _dsc
        finally:
            pipe.execute()
            pipe.reset()

    def _get_control_data(self, ppr):
        """Obtain essential data for policy decisionmaking

        This is the routine which keeps track of emails in Redis.  It
        combines all of its requests into a single pipelined (atomic)
        transaction.  When counting recipients, the record is a string
        consisting of the timestamp and the recipient serial number
        separated by a colon, in order to ensure that each recipient
        is listed as an attempt in the log.  The score is always the
        floating-point return value of time.time()
        """
        # cache the current(ish) time
        time_now = time.time()
        time_now_s = str(time_now)

        user = ppr.user

        # create a dict for Redis.zadd()
        if self.counting_recipients:
            tries_dict = {
                (time_now_s + f":{i:05d}"): time_now
                for i, r in enumerate(ppr.recipients)
            }
        else:
            tries_dict = {time_now_s: time_now}
        # set up the Redis keys
        tries_key = self._fmtkey(user, "attempts")
        limit_key = self._fmtkey(user, "limit")
        margin_key = self._fmtkey(user, "margin")
        # Create a Redis pipeline to atomize a set of instructions
        pipe = self.redis.pipeline()
        # Clear the list down to just the last interval seconds, generally a day
        pipe.zremrangebyscore(tries_key, 0, time_now - float(self.interval))
        # Add this/these try(es)
        pipe.zadd(tries_key, tries_dict)
        # Get control data: the limit, the margin, the attempts list
        pipe.get(limit_key)
        pipe.get(margin_key)
        pipe.zrange(tries_key, 0, -1)
        # Set expires on all this stuff so that if they don't send
        # email for a day, their data won't still be sitting around
        # takin' up space; these return nil values so we have to
        # ignore them when we get the results
        pipe.expire(tries_key, self.interval)
        pipe.expire(limit_key, self.interval)
        pipe.expire(margin_key, self.interval)
        # Do the thing!
        results = pipe.execute()
        # The non-retrieval operations still have return values, so we ignore them
        removed, _, limit, margin, attempts, _, _, _ = results
        # Always polite to reset your pipe
        pipe.reset()
        # Attempt typecasting on the margin number, which is allowed
        # to be either int or float
        m = self._cast_margin(margin)
        # If no limit is defined, that means there is no quota profile
        # for the user, and we just return None here; we must test against
        # None because it might be 0
        return (int(limit) if limit is not None else None, m, attempts)

    def _cast_margin(self, margin_bytes):
        """Convenience method

        :param bytes margin_bytes: margin value from Redis

        Get the correct type (int or float) from the provided bytestring.

        """
        try:
            m = int(margin_bytes)
        except:
            try:
                m = float(margin_bytes)
            except:
                m = 0
        return m

    def current_quota(
        self, user: str, quota: Optional[Quota] = None
    ) -> Tuple[int, List[str]]:
        """Provide real-time remaining quota for a user

        :param str user: user-identifier

        :param chapps.models.Quota quota: optional
          :class:`~chapps.models.Quota` record

        :returns: (*remaining quota count*, [*remarks*,...])

        :rtype: Tuple[int, List[str]]

        The caller is anticipated to be the API.  The **User** and **Quota**
        are both available, so the **Quota** may be provided, but it is not
        required.  The `user` parameter is expected to contain a string at
        present, though this may change as the :mod:`pydantic` data models
        become more tightly integrated into the codebase.

        The return value, intended for wrapping by the API and transmission to a client, is a tuple composed of:

          1. the number of transmission attempts remaining to the user at the
             moment the query executed

          2. a list of remarks created by the inspection routine

        """
        attempts_key = self._fmtkey(user, "attempts")
        limit_key = self._fmtkey(user, "limit")
        pipe = self.redis.pipeline()
        pipe.zremrangebyscore(
            attempts_key, 0, time.time() - float(self.interval)
        )
        pipe.get(limit_key)
        pipe.zrange(attempts_key, 0, -1)
        results = pipe.execute()
        _, limit_bytes, attempts_bytes = results
        pipe.reset()
        limit = (
            int(limit_bytes)
            if limit_bytes is not None
            else quota.quota
            if quota is not None
            else None
        )
        response = limit - len(attempts_bytes) if limit else 0
        remarks = []
        if attempts_bytes:
            last = attempts_bytes[-1]
            try:
                last = float(last)
            except ValueError:
                last = float(last.split(b":")[0])
            last_try = time.strftime(TIME_FORMAT, time.gmtime(last))
            remarks.append(f"Last send attempt was at {last_try}")
        if not limit_bytes:
            remarks.append(f"There is no cached quota limit for {user}.")
        if not quota:
            remarks.append(f"There is no quota configured for user {user}.")
        if not limit:
            remarks.append(
                f"No limit could be found; returning zero xmits remaining."
            )
        return (response, remarks)

    def reset_quota(self, user: str) -> Tuple[int, List[str]]:
        """Reset quota for user

        :param str user: user-identifier

        :returns: (*number of records dropped*, [*remarks*,...])

        :rtype: Tuple[int,List[str]]

        This method is intended for real-time management of the Redis
        configuration mirror.  It will drop all the attempts from the
        outbound-quota transmission-tracking list for the named user.

        """
        attempts_key = self._fmtkey(user, "attempts")
        pipe = self.redis.pipeline()
        pipe.zrange(attempts_key, 0, -1)
        pipe.delete(attempts_key)
        results = pipe.execute()
        pipe.reset()
        attempts = results[0]
        if attempts:
            msg = f"Attempts (quota) reset for {user}:"
            n_att = len(attempts)
        else:
            n_att = 0
            msg = f"No attempts to reset for {user}:"
        msg += f" {n_att} xmits dropped"
        return (n_att, [msg])

    def refresh_policy_cache(self, user: str, quota: Quota):
        """API adapter method for refreshing policy config cache"""
        self.acquire_policy_for(user, quota.quota)
        return self.current_quota(user, quota)

    def _store_control_data(self, user, quota, margin=0):
        """Using a context manager, build up a set of instructions to store control data"""
        with self._control_data_storage_context() as dsc:
            if type(margin) == float:
                if margin > 1.0:
                    if margin < 100.0:
                        margin = margin / 100.0
                    else:
                        raise TypeError(
                            "margin must be a positive integer or a positive float less than 1 (a percentage)"
                        )  # pragma: no cover
                margin = int(margin * float(quota))
            dsc(user, quota, margin)

    def _detect_control_data(self, user):
        """See if there is control data in Redis for a particular sender"""
        key = self._fmtkey(user, "limit")
        res = None
        try:
            res = self.redis.get(key)
        except redis.exceptions.ResponseError:  # pragma: no cover
            # sometimes a key which should not exist still shows up,
            # and accessing them causes this error.  we choose to
            # pretend nothing happened and just delete the key
            self.redis.delete(key)
        return res

    def acquire_policy_for(self, user: str, quota: Optional[int] = None):
        """Populate Redis with policy config data for a user

        :param str user: user-identifier

        :param int quota: optional quota to load for the user.  This is
          provided mainly to optimize actions taken by the API.

        Go get the policy for a sender from the policy adapter.

        If the margin needs to be configured on a per-sender basis, this is the
        place to adjust that.  Right now, the margin is set in the config file,
        and applied to each user as policy config is loaded.

        """
        if not quota:
            with self._adapter_handle() as adapter:
                quota = adapter.quota_for_user(user)
        if quota:
            self._store_control_data(user, quota, self.params.margin)

    def _approve_policy_request(self, ppr: OutboundPPR):
        """Determine whether this email falls within the quota

        :param chapps.outbound.OutboundPPR ppr: the Postfix payload

        Returns True if this email is within the quota.

        This routine implements memoization on `ppr.instance` in order to
        overcome the Postfix double-checking weirdness.  Sometimes, Postfix
        sends a request about a given email twice, but this is easy to spot
        because they will have the same value for `ppr.instance`.

        :meta public:
        """
        user = ppr.user
        if not self._detect_control_data(user):
            self.acquire_policy_for(user)
        return self._evaluate_policy_request(ppr)

    def _get_delta(self, ppr, attempts):
        """Obtain the number of seconds between successive attempts

        This routine should calculate the number of seconds between this
        attempt and the previous.  To do so, it must take into account whether
        we are counting each recipient, and parse the attempt record
        accordingly.  When counting recipients, the record is a string
        consisting of the timestamp and the recipient serial number separated
        by a colon, in order to ensure that each recipient is listed as an
        attempt in the log.

        .. admonition: Experimental

          There is some subtle problem with either or both of the logic here
          and in the tests, and so since it was initially provided as an
          interesting extra feature, it is currently disabled and considered
          experimental.  At some point, I intend to come back to it.

        """
        if len(attempts) < 2:
            return float("inf")
        delta_index = [-1, -2]
        if not self.counting_recipients:
            if len(attempts) < 2:
                return float("inf")
        elif len(attempts) > len(ppr.recipients):
            # skip back all but one recipient
            recipients_offset = 0 - len(ppr.recipients)
            delta_index = [d + recipients_offset for d in delta_index]
        else:
            return float("inf")  # automatically wins
        logger.debug(
            f"Looking at time-delta for {ppr}: indices {delta_index!r}"
        )
        try:
            timestamps = [
                float(t.decode("utf-8").split(":")[0])
                if ":" in t.decode("utf-8")
                else float(t.decode("utf-8"))
                for t in [
                    attempts[i] for i in delta_index if i < len(attempts)
                ]
            ]
        except IndexError:
            msg = (
                f"Recipients={-recipients_offset}"
                f" delta_indices={delta_index!r}"
                f" Attempts: (#{len(attempts)})"
            )
            if len(attempts) < 10:
                msg += f" {attempts!r}"
            logger.exception(msg)
            return float("inf")
        if len(timestamps) == 2:
            logger.debug(
                f"attempts: {[attempts[i] for i in delta_index]}; timestamps: {timestamps!r}"
            )
            return timestamps[0] - timestamps[1]
        return float("inf")  # return a large value

    def _evaluate_policy_request(self, ppr):
        """This actually checks to see if it's okay to send the email.

        .. todo::

          in this routine, it would be possible to send pub/sub messages via
          Redis to consumers who might be interested to know that a particular
          user's send-attempts list is over a certain length

        """
        instance, user = ppr.instance, ppr.user
        try:  # this may raise TypeError if the user is unknown
            limit, margin, attempts = self._get_control_data(ppr)
        except Exception:  # pragma: no cover
            logger.exception("UNEXPECTED")
            logger.debug(
                f"Returning denial indicator for {instance} (unexpected exception)."
            )
            return False
        if not limit:  # user does not have a quota profile
            return False
        if len(attempts) < 2:  # this is the first attempt in the Redis history
            logger.debug(f"Returning OK for {instance} (first attempt).")
            return True
        if self.min_delta != 0:  # set up for checking on throttle
            logger.debug(
                f"Checking throttle: {ppr.user}:{instance} limit: {limit} margin: {margin} tries: {len(attempts)}"
            )
            this_delta = self._get_delta(ppr, attempts)
            if this_delta < float(self.min_delta):  # trying too fast
                logger.debug(
                    f"Rejecting {instance} of {user} for trying too fast. ({this_delta}s since last attempt)"
                )
                return False
        if (
            len(attempts) > limit
        ):  # not too fast, check how many send attempts on record
            ### TODO: alert when the attempts list is really long -- perhaps via Redis pub/sub
            if (
                len(attempts) - margin > limit
                or len(attempts) - len(ppr.recipients) >= limit
            ):
                logger.debug(
                    f"Rejecting {instance} of {user} for having too many attempts in the last interval: recip: {len(ppr.recipients)} limit: {limit}; tries: {len(attempts)}"
                )
                return False
            else:
                logger.debug(
                    f"Returning OK for {instance} recip: {len(ppr.recipients)} limit: {limit}; tries: {len(attempts)} (within margin)."
                )
        logger.debug(f"Returning OK for {instance} (under quota).")
        return True


class SenderDomainAuthPolicy(EmailPolicy):
    """Policy manager implementing domain and whole-email matching for senders

    This class encapsulates explicit policy regarding what sorts of
    masquerading authenticated users are allowed to do.  Currently, two sorts
    of matches are handled, in succession.

    First, the domain part of the email address, the entire string after the
    `@`, is matched against **Domain** entries linked to the **User**.

    If there is no **Domain** match, then **Email** entries linked to the
    **User** are checked.  **Email** entries must match the entirety of a
    policy request's `sender` attribute in order to pass.

    """

    adapter_class = SDAAdapter
    redis_key_prefix = "sda"
    """Sender domain auth Redis key prefix"""
    # initialization is when we plug in the config
    def __init__(self, cfg: CHAPPSConfig = None):
        """Set up a new sender domain authorization policy manager

        :param chapps.config.CHAPPSConfig cfg: optional config override

        """
        super().__init__(cfg)  # sets attrs 'config' and 'redis'

    # every subclass has one of these, with a unique name, and fine
    # but maybe there should also be a generic single entry point
    def sender_domain_key(self, ppr: OutboundPPR) -> str:
        """Create a Redis key for a user->domain mapping

        :param chapps.outbound.OutboundPPR ppr: a Postfix payload

        :returns: the sender domain key, by obtaining the domain part of the
          email address from `ppr.sender`

        :rtype: str

        """
        return self._sender_domain_key(ppr.user, self._get_sender_domain(ppr))

    def sender_email_key(self, ppr) -> str:
        """Create a Redis key for a user->email mapping

        :param chapps.outbound.OutboundPPR ppr: a Postfix payload

        :returns: the sender email key, by obtaining the email address from
          `ppr.sender`

        :rtype: str

        """
        return self._sender_domain_key(ppr.user, ppr.sender)

    # factored out for use in API
    def _sender_domain_key(self, user: str, domain: str) -> str:
        """Passes its two string params to _fmtkey

        :meta public:
        :param str user: user-identifier

        :param str domain: origin domain or email address

        :returns: a Redis key

        :rtype: str

        Should be called `_sender_auth_key` since it works with both domains
        and email addresses.

        """
        return self._fmtkey(user, domain)

    # determine the domain of the sender address, if any
    @functools.lru_cache(maxsize=2)
    def _get_sender_domain(self, ppr: OutboundPPR) -> str:
        """Returns the domain portion of `ppr.sender`

        :param chapps.outbound.OutboundPPR ppr: a Postfix payload

        :returns: the domain part of `ppr.sender`

        :rtype: str

        :raise chapps.signals.TooManyAtsException: if there are more than one
          `@` in `ppr.sender`

        :raise chapps.signals.NotAnEmailAddressException: if there is no `@`
          in `ppr.sender`

        :raise chapps.signals.NullSenderException: if `ppr.sender` is
          `None`

        :meta public:

        """
        if ppr.sender:
            return ppr.domain_from(ppr.sender)
        raise NullSenderException

    # We will need to be able to access policy data in the RDBMS
    def _detect_control_data(self, user, domain):
        """Look for SDA control data for a user"""
        key = self._sender_domain_key(user, domain)
        res = None
        try:
            res = self.redis.get(key)
        except redis.exceptions.ResponseError:  # pragma: no cover
            self.redis.delete(key)
        # logger.debug(f"Found {key} = {res!r} in Redis")
        return res if res is None else int(res)

    def _get_control_data(self, ppr):
        """Cascade through control data searches: domain, email"""
        return self._detect_control_data(
            ppr.user, self._get_sender_domain(ppr)
        ) or self._detect_control_data(ppr.user, ppr.sender)

    # We will need to be able to store data in Redis
    def _store_control_data(self, ppr: OutboundPPR, allowed: int):
        """Stuff control data into Redis for domain auth"""
        # logger.debug(
        #     f"store request: {ppr.user} {self._get_sender_domain(ppr)}"
        #     f" {allowed!r}"
        # )
        with self._control_data_storage_context() as dsc:
            dsc(ppr.user, self._get_sender_domain(ppr), allowed)

    def _store_email_control_data(self, ppr, allowed):
        """Stuff control data into Redis for email auth"""
        # logger.debug(f"store request: {ppr.user} {ppr.sender} {allowed!r}")
        with self._control_data_storage_context() as dsc:
            dsc(ppr.user, ppr.sender, allowed)

    # We will need a Redis storage context manager in order to mimic
    # the structure of OQP -- metaprogramming opportunity
    @contextmanager
    def _control_data_storage_context(
        self, expire_time: int = seconds_per_day
    ):
        """Context manager for storing SDA policy cache data in Redis"""
        pipe = self.redis.pipeline()
        fmtkey = self._fmtkey

        def _dsc(user, domain, allowed):
            key = fmtkey(user, domain)
            # logger.debug(f"Storing {key} = {allowed} in Redis")
            pipe.set(key, allowed, ex=seconds_per_day)

        try:
            yield _dsc
        finally:
            pipe.execute()
            pipe.reset()

    # How to obtain control data
    def acquire_policy_for(self, ppr) -> bool:
        """Populate Redis with policy config

        :param chapps.outbound.OutboundPPR ppr: a Postfix payload

        :returns: whether the policy allows `ppr`

        :rtype: bool

        Populates Redis and return the policy result for `ppr`.

        """
        # logger.debug(f"acq pol for {ppr!r}")
        with self._adapter_handle() as adapter:
            allowed = adapter.check_domain_for_user(
                ppr.user, self._get_sender_domain(ppr)
            )
            self._store_control_data(ppr, 1 if allowed else 0)
            # logger.debug(
            #     f"RDBMS: policy {allowed!r} for {ppr.user} from domain"
            #     f" {self._get_sender_domain(ppr)}"
            # )
            if not allowed:  # domain not allowed, check email
                allowed = adapter.check_email_for_user(ppr.user, ppr.sender)
                if allowed is not None:
                    self._store_email_control_data(ppr, 1 if allowed else 0)
                    # logger.debug(
                    #     f"RDBMS: policy {allowed!r} for {ppr.user} as"
                    #     f" {ppr.sender}"
                    # )
        return allowed

    # This is the main purpose of the class, to answer this question
    def _approve_policy_request(self, ppr: OutboundPPR) -> bool:
        """Returns true if `ppr` represents an authorized email

        :param chapps.outbound.OutboundPPR ppr: a Postfix payload

        :returns: whether the email represented by `ppr` should be
          transmitted

        :rtype: bool

        Given a PPR, say whether this user is allowed to send as the
        apparent sender domain or address.  Memoize result in the
        instance cache.

        """
        result = self._get_control_data(ppr)
        if result is None:
            result = self.acquire_policy_for(ppr)
            # logger.debug(f"Obtained {result!r} from RDBMS.")
        else:
            # logger.debug(f"Returning {result!r} from Redis.")
            pass
        return bool(int(result))

    def _decode_policy_cache(self, result) -> SDAStatus:
        """Decode the {None, 0, 1} response from Redis into an Enum

        :params int result: b'0', b'1', or None

        :returns: an SDAStatus corresponding to `result`

        :rtype: chapps.models.SDAStatus

        The :class:`Enum` :class:`~chapps.models.SDAStatus` represents
        these results as nonexistent, prohibited or authorized respectively.

        """
        if result is not None:
            result = int(result)
            if result:
                return SDAStatus.AUTH
            else:
                return SDAStatus.PROH
        else:
            return SDAStatus.NONE

    # For the API -- inspect state
    def check_policy_cache(self, user: str, domain: str) -> SDAStatus:
        """Check a particular policy cache entry for the API

        :param str user: user-identifier

        :param str domain: domain or email to check

        :returns: the cached policy

        :rtype: chapps.models.SDAStatus

        """
        return self._decode_policy_cache(
            self._detect_control_data(user, domain)
        )

    def clear_policy_cache(self, user: str, domain: str) -> SDAStatus:
        """Clear a specific policy cache entry

        :param str user: user-identifier

        :param str domain: domain or email to clear

        :returns: the previous policy

        :rtype: SDAStatus

        """
        prev = self.check_policy_cache(user, domain)
        if prev != SDAStatus.NONE:
            self.redis.delete(self._sender_domain_key(user, domain))
        return prev

    def bulk_clear_policy_cache(
        self,
        users: List[str],
        domains: List[str] = None,
        emails: List[str] = None,
    ):
        r"""Clear SDA policy cache
        for **User**\ s x [**Domain**\ s + **User**\ s]

        :param List[str] users: a list of user-identifiers
        :param Optional[List[str]] domains: a list of domain names
        :param Optional[List[str]] emails: a list of email addresses
        :rtype: None

        """
        # there seems to be no max pipeline size
        # but if things get sketchy, we can chunk this
        domains = domains or []
        emails = emails or []
        with self.redis.pipeline() as pipe:
            for d in domains + emails:
                for u in users:
                    # logger.debug(f"bulk_clear: erasing {u}:{d}")
                    pipe.delete(self._sender_domain_key(u, d))
            pipe.execute()
            pipe.reset()

    def bulk_check_policy_cache(
        self,
        users: List[str],
        domains: List[str] = None,
        emails: List[str] = None,
    ) -> Dict[str, Dict[str, SDAStatus]]:
        """Map auth subject onto user status

        :param List[str] users: a list of user-identifiers
        :param Optional[List[str]] domains: a list of domain names
        :param Optional[List[str]] emails: a list of email addresses

        :returns: an auth subject => user => status map as described below

        :rtype: Dict[str, Dict[str, SDAStatus]]

        Builds a map keyed on auth subject (**Domain** and/or **Email**), full
        of maps from username to status.  It looks a bit like this:

        .. code::python

            bulk_check_result = {
                'example.com': { 'user@example.com': SDAStatus.AUTH,
                                 'terminated@example.com': SDAStatus.PROH,
                },
                'chapps.io': { 'user@example.com': SDAStatus.NONE,
                               'terminated@example.com': SDAStatus.NONE,
                }
            }

        Mainly intended for use by the API.

        """
        emails = emails or []
        domains = domains or []
        with self.redis.pipeline() as pipe:
            for d in domains + emails:
                for u in users:
                    # logger.debug(f"bcpc seeking SDA for {u} from {d}")
                    pipe.get(self._sender_domain_key(u, d))
            results = deque(pipe.execute())
            pipe.reset()
        # logger.debug(f"bcpc results: {results!r}")
        return {
            d: {u: self._decode_policy_cache(results.popleft()) for u in users}
            for d in domains + emails
        }



"""
Operational Configuration
-------------------------

CHAPPS configures itself at the library level.  When it is first launched,
the library will create a config file for itself if it does not find one
at its default config path, `/etc/chapps/chapps.ini`, or the value of
the environment variable `CHAPPS_CONFIG` if it is set.  When it does,
default settings for all available submodules will be produced.

Any instance of CHAPPS requires the general CHAPPS settings, adapter settings,
and the Redis settings.  These control basic features of CHAPPS and tell it
how to access its brains.

Each service script most likely runs a unique policy handler.  If only one
service is being used, only the settings for the policies of that handler will
be needed, plus the ones mentioned above.

Policy handlers (from :py:mod:`chapps.switchboard`) take their listener
configuration from that of their related policy.  Each may each be configured
to use separate listening addresses and ports, so that they may run
simultaneously on the same server.

.. note::

    For multi-policy handlers, the settings used are taken from the first
    handler found to have config elements named `listen_address` and
    `listen_port`.  It is recommended to configure those elements only on one
    active policy, or to keep them in sync on all policies which are handled
    together.

"""
import collections.abc
import configparser
import dns.resolver
import dns.exception
from functools import cached_property
from pathlib import Path
from os import environ as env
from chapps.util import AttrDict, VenvDetector
from chapps._version import __version__
from chapps.signals import (
    HELOWLException,
    AddressDoesNotMatchDNS,
    NameDoesNotMatchPTR,
)
import logging
from typing import Union

try:
    from dotenv import load_dotenv

    load_dotenv()
except Exception:
    pass

logger = logging.getLogger(__name__)


class CHAPPSConfig:
    """The configuation object

    Mostly a wrapper around :py:mod:`configparser`, with the most
    commonly used portions wrapped in :py:mod:`chapps.util.AttrDict`

    """

    configs = dict()

    @classmethod
    def get_config(cls):
        cfg_fname = str(cls.what_config_file(VenvDetector().confpath))
        if len(cls.configs) and cfg_fname in cls.configs:
            return cls.configs[cfg_fname]
        new_config = CHAPPSConfig()
        cls.configs[cfg_fname] = new_config
        return new_config

    # ultimately, we may need also to allow for a command-line option
    @staticmethod
    def what_config_file(
        default_pathname: str = "/etc/chapps/chapps.ini",
    ) -> Path:
        """Determine what config file to read.

        This is to allow for easier addition of a command-line option.
        Also encapsulates search for possible file pointed to by the
        environment setting `CHAPPS_CONFIG`

        """
        config_file = Path(env.get("CHAPPS_CONFIG", default_pathname))
        return config_file

    @staticmethod
    def setup_config(
        cp: configparser.ConfigParser,
    ) -> configparser.ConfigParser:
        """Setup default config pattern on the parser passed in

        :param configparser.ConfigParser cp: a
           :py:class:`configparser.ConfigParser` instance to hold the
           default config

        This routine establishes the default configuration.  It returns
        the same object which was passed to it.
        """
        cp["CHAPPS"] = {
            "payload_encoding": "utf-8",
            "user_key": "sasl_username",
            "require_user_key": True,
            "no_user_key_response": "REJECT Rejected - Authentication failed",
            "password": (
                "effda33d276c1d5649f3933a6d6b286e"
                "d7eaaede0b944221e7699553ce0558e2"
            ),
            "listener_backlog": 100,
        }
        cp["PolicyConfigAdapter"] = {
            "adapter": "mariadb",
            "db_host": "localhost",
            "db_port": "3306",
            "db_name": "chapps",
            "db_user": "chapps",
            "db_pass": "chapps",
        }
        cp["Redis"] = {
            "sentinel_servers": "",
            "sentinel_dataset": "",
            "server": "localhost",
            "port": "6379",
        }
        cp["OutboundQuotaPolicy"] = {
            "listen_address": "localhost",
            "listen_port": 10225,
            "margin": 0.10,
            "min_delta": 0,
            "counting_recipients": True,
            "rejection_message": "REJECT Rejected - outbound quota fulfilled",
            "acceptance_message": "DUNNO",
            "null_sender_ok": False,
        }
        cp["GreylistingPolicy"] = {
            "listen_address": "localhost",
            "listen_port": 10226,
            "rejection_message": (
                "DEFER_IF_PERMIT Service temporarily"
                " unavailable - greylisted"
            ),
            "acceptance_message": "DUNNO",
            "null_sender_ok": False,
            "whitelist_threshold": 10,
        }
        cp["SPFEnforcementPolicy"] = {
            "listen_address": "localhost",
            "listen_port": 10227,
            "whitelist": [],
            "null_sender_ok": False,
            "spf_query_timeout": 20,
        }
        cp["PostfixSPFActions"] = {
            "passing": "prepend",
            "fail": "550 5.7.1 SPF check failed: {reason}",
            "temperror": "451 4.4.3 SPF record(s) temporarily unavailable: {reason}",
            "permerror": "550 5.5.2 SPF record(s) are malformed: {reason}",
            "none_neutral": "greylist",
            "softfail": "greylist",
        }
        cp["SenderDomainAuthPolicy"] = {
            "listen_address": "localhost",
            "listen_port": 10225,
            "rejection_message": "REJECT Rejected - not allowed to send mail from this domain",
            "acceptance_message": "DUNNO",
            "null_sender_ok": False,
        }
        return cp

    @staticmethod
    def write_config(cp, fn) -> Path:
        """Write the ConfigParser contents to disk.

        :param configparser.ConfigParser cp: a ConfigParser object
        :param Union[str, pathlib.Path] fn: path of the config file to write

        If the location's parent directory does not exist, CHAPPS
        will attempt to create it.  If CHAPPS can open the file, it
        writes the contents of `cp` into the file referred to by `fn`.

        Returns a :class:`pathlib.Path` which points at the newly-written file.

        """
        config_file = Path(fn)
        if not config_file.parent.exists():
            try:  # attempt to make any missing parent directories
                config_file.parent.mkdir(0o777, True)
            except OSError as e:
                logger.error(
                    "The specified config file's directory did not exist and"
                    f" could not be created.  File: {str(config_file)}"
                )
                raise e  # possibly this should not be re-raised
        with config_file.open("w") as fh:
            cp.write(fh)
        return config_file

    def __init__(self):
        """Setup a new CHAPPSConfig instance

        This routine does a bunch of different setup to provide more than just
        the on-disk configuration to the running instance. It provides the path
        to the config file that was used to configure the session.  It also
        provides a symbol to refer to the version number.

        It creates a :mod:`chapps.util.VenvDetector` in order to set up the
        path to the README-API.md file.  This could also be used to change the
        default config location when running in a `venv`.  That would
        eliminate the need for prefacing commands with a phrase setting the
        config-file location.

        It causes a config file full of defaults to be written to disk if it
        does not find a file to read.  If it does find a file, it uses the
        settings from that file to overlay the defaults already set up on the
        config object.  It is for this reason that an API method is provided to
        refresh the file on disk with any new settings which might have been
        introduced since the software was last configured.

        """
        # Create and initialize the config
        self.venvdetector = VenvDetector()
        config_file = CHAPPSConfig.what_config_file(self.venvdetector.confpath)
        self.configparser = configparser.ConfigParser(interpolation=None)
        CHAPPSConfig.setup_config(self.configparser)

        # if this is a real run and not a doc build...
        if not self.venvdetector.sb:
            # Initialize a config file if none
            if not config_file.exists():
                logger.debug("Writing new config file " + str(config_file))
                CHAPPSConfig.write_config(self.configparser, config_file)
            else:
                logger.debug("Reading from config file " + str(config_file))
                self.configparser.read(str(config_file))
        self.configparser["CHAPPS"]["config_file"] = str(config_file)
        self.configparser["CHAPPS"]["version"] = f"CHAPPS v{__version__}"
        self.configparser["CHAPPS"]["docpath"] = str(self.venvdetector.docpath)
        self.chapps = AttrDict(self.configparser["CHAPPS"])
        self.adapter = AttrDict(self.configparser["PolicyConfigAdapter"])
        self.actions_spf = AttrDict(self.configparser["PostfixSPFActions"])
        self.redis = AttrDict(self.configparser["Redis"])
        # these are somewhat obsolete now
        self.policy_oqp = AttrDict(self.configparser["OutboundQuotaPolicy"])
        self.policy_sda = AttrDict(self.configparser["SenderDomainAuthPolicy"])
        self.policy_grl = AttrDict(self.configparser["GreylistingPolicy"])
        self.policy_spf = AttrDict(self.configparser["SPFEnforcementPolicy"])
        if not self.venvdetector.sb:
            logger.debug("Returning config built from " + str(config_file))

    def get_block(self, blockname) -> AttrDict:
        """Attempt to get a top-level block of the config as an AttrDict.

        :param str blockname: the name of the block

        Return `None` if it cannot be found.

        """
        try:
            return AttrDict(self.configparser[blockname])
        except Exception:
            pass
        return None

    def write(self, location: Union[str, Path] = None):
        """Write the current config to disk.

        :param Union[str,pathlib.Path] location: where to write the file

        Currently manages exceptions for elements of the `[CHAPPS]` section of
        the config, for parameters which are not specified in the file.  It
        removes them from the config, writes the file, then restores their
        values.  These are:

          * the config file itself (as a :class:`Path`)
          * the CHAPPS version
          * the READMEs' directory (as a :class:`Path`)

        """

        location = Path(location or self.chapps.config_file)
        config_file = self.chapps.config_file
        version = self.chapps.version
        docpath = self.chapps.docpath
        self.configparser.remove_option("CHAPPS", "config_file")
        self.configparser.remove_option("CHAPPS", "version")
        self.configparser.remove_option("CHAPPS", "docpath")
        result = CHAPPSConfig.write_config(self.configparser, location)
        if config_file:
            self.configparser["CHAPPS"]["config_file"] = config_file
        self.configparser["CHAPPS"]["version"] = version
        self.configparser["CHAPPS"]["docpath"] = docpath
        return result

    @cached_property
    def helo_whitelist(self):
        if "helo_whitelist" not in self.chapps:
            return {}
        return {
            k: v for k, v in self._parse_whitelist(self.chapps.helo_whitelist)
        }

    @staticmethod
    def _parse_whitelist(speclist):
        loopback = "127.0.0.1"
        for spec in speclist.split(";"):
            name, src_ip = spec, None
            if ":" in spec:
                name, src_ip = spec.split(":", 1)
            if src_ip and src_ip == loopback:
                yield (name, src_ip)
                continue
            try:
                lookup = dns.resolver.resolve(name)
                dns_ip = str(lookup[0])
                if src_ip:
                    if src_ip != dns_ip:
                        raise AddressDoesNotMatchDNS(
                            f"source {src_ip} != {dns_ip} from DNS for {name}"
                        )
                else:
                    src_ip = dns_ip
                ptr = dns.resolver.resolve_address(src_ip)
                ptr_name = str(ptr[0])
                if name + "." != ptr_name:
                    raise NameDoesNotMatchPTR(
                        f"name {name}. != {ptr_name} "
                        f"from DNS (PTR for {src_ip})"
                    )
                yield (name, src_ip)
            except HELOWLException as e:
                logger.exception(e)
                pass
            except dns.exception.DNSException as e:
                logger.exception(e)
                pass



"""
Route factories and other reusable code
---------------------------------------

Factories for utility functions and API routes
are defined in this module
along with some `FastAPI`_ dependencies.

The route factories perform the repetitive grunt work required to set up the
typical 'create', 'read', 'update', 'delete' and 'list' functions needed for
basic object management.  In order to avoid extra levels of metaprogramming,
the parameter name for the record ID of the main object involved in a
factory-generated API call is ``item_id``, since it is clear, brief and
generic.  Apologies to future subclassors who want an 'items' table.

These route factories are used to create all the routes for
:mod:`~chapps.rest.routers.users`, :mod:`~chapps.rest.routers.emails`,
:mod:`~chapps.rest.routers.domains`, and :mod:`~chapps.rest.routers.quotas`.

"""
from typing import Optional, List
from sqlalchemy.orm import sessionmaker
from sqlalchemy.exc import IntegrityError, OperationalError
from fastapi import status, Depends, Body, HTTPException
from functools import wraps
import inspect
import logging
from chapps.dbsession import sql_engine
from chapps.models import (
    CHAPPSModel,
    CHAPPSResponse,
    AssocOperation,
    DeleteResp,
    TextResp,
)
from chapps.dbmodels import JoinAssoc
import chapps.logging

logger = logging.getLogger(__name__)
logger.setLevel(chapps.logging.DEFAULT_LEVEL)


async def list_query_params(
    skip: Optional[int] = 0,
    limit: Optional[int] = 1000,
    q: Optional[str] = "%",
) -> dict:
    """FastAPI dependency for list queries"""
    return dict(q=q, skip=skip, limit=limit)


def model_name(cls) -> str:
    """Convenience function to get the lowercase name of a model"""
    return cls.__name__.lower()


def load_model_with_assoc(cls, assoc: List[JoinAssoc], engine=sql_engine):
    """Create a closure which loads an object along with arbitrary associations

    This isn't meant to create an API route on its own, but it may be used in
    API routes.  It is mainly used in :mod:`~chapps.rest.routers.live` routes,
    which are currently all one-offs, not created by factories.  In order to
    return a closure which can work in any context, it does not return a
    coroutine but a standard synchronous closure.

    :param ~.CHAPPSModel cls: a data model class
    :param assoc: a list of associations (as
      :class:`~.JoinAssoc` objects)
    :param Optional[~sqlalchemy.engine.Engine] engine: defaults to
      :const:`chapps.dbsession.sql_engine` if not specified
    :rtype: callable
    :returns: a closure as follows:

      .. py:function:: f(item_id: int, name: Optional[str])

        :param int item_id: if non-zero, the ID of the main record
        :param Optional[str] name: if `item_id` is 0, the `name` of the record
          to match.
        :rtype: Tuple[~.CHAPPSModel, Dict[str, List[~.CHAPPSModel]], List[str]]
        :returns: a :obj:`tuple` containing:

          1. the object loaded by ID or name

          2. that object's associations in a :obj:`dict` keyed on attribute
             name (e.g. 'quota', 'domains')

          3. a list of string remarks, which may have no contents

    """
    mname = model_name(cls)
    assoc_s = "_".join([a.assoc_name for a in assoc])
    fname = f"load_{mname}_with_{assoc_s}"
    # Session = sessionmaker(engine)

    @db_wrapper(cls=cls, engine=engine)
    def get_model_and_assoc(item_id: int, name: Optional[str]):
        remarks = []
        items = {k: None for k in [mname, *[a.assoc_name for a in assoc]]}
        # session is a global provided by the decorator
        if item_id:
            items[mname] = session.scalar(cls.select_by_id(item_id))
        if name and not items[mname]:
            items[mname] = session.scalar(cls.select_by_name(name))
            if items[mname] and item_id:
                remarks.append(
                    f"Selecting {mname} {items[mname].name} with "
                    f"id {items[mname].id} by name because "
                    "provided id does not exist."
                )
        if not items[mname] and not (item_id or name):
            logger.debug(  # log this, as it is weird
                f"{fname}({item_id!r}, {name!r}): unable to load {mname}"
            )
            raise HTTPException(  # describe error to the caller
                status_code=status.HTTP_400_BAD_REQUEST,
                detail=(
                    f"One of {cls.id_name()} or name must be provided. "
                    "If both are provided, {cls.id_name()} is preferred."
                ),
            )
        if items[mname]:
            for a in assoc:
                items[a.assoc_name] = getattr(items[mname], a.assoc_name)
        else:
            detail = "No {mname} could be found with "
            if item_id:
                detail += "id {item_id}"
                if name:
                    detail += f" or with "
            if name:
                detail += "name {name}"
            raise HTTPException(
                status_code=status.HTTP_404_NOT_FOUND, detail=detail
            )
        return (items.pop(mname), items, remarks)

    get_model_and_assoc.__name__ = fname
    return get_model_and_assoc


def load_models_with_assoc(
    cls: CHAPPSModel, *, assoc: JoinAssoc, engine=sql_engine
) -> callable:
    """Build a map of source name => associated object id

    :param cls: source model

    :param assoc: a join association representing the associated model

    :param engine: override the SQLA engine if desired

    :returns: a mapper function which accepts a list of IDs of the source model
      and returns a list of dicts with `<source_model>_name` and
      `<assoc_model>_id` fields, mapping the source objects onto the IDs of
      their associated objects of the configured type.

    .. todo::

      Employ eager loading on the target association.

    """
    mname = model_name(cls)
    fname = f"eager_load_{mname}_with_{assoc.assoc_name}"
    # Session = sessionmaker(engine)

    @db_wrapper(cls=cls, engine=engine)
    def map_model_names_to_assoc(
        item_ids: List[int],  # name_tail: Optional[str] = None
    ):
        eager_loaded_models = session.scalars(
            cls.select_by_ids(
                item_ids, getattr(cls.Meta.orm_model, assoc.assoc_name)
            )
        )
        return list(eager_loaded_models)

    map_model_names_to_assoc.__name__ = fname
    return map_model_names_to_assoc


def db_wrapper(  # a decorator with parameters
    *,
    cls,
    engine=sql_engine,
    exception_message: str = ("{route_name}:{model}"),
    empty_set_message: str = ("Unable to find a matching {model}"),
):
    """Decorator for database interactions

    :param ~chapps.models.CHAPPSModel cls: the data model class

    :param ~sqlalchemy.engine.Engine engine: an :mod:`SQLAlchemy` engine, which
      defaults to the package-wide one declared in
      :mod:`~chapps.dbsession`

    :param str exception_message: a message to include if any untrapped
      exception occurs; defaults to ``{route_name}:{model}``.  Only those two
      symbols are available for expansion.  All arguments are appended.

    :param str empty_set_message: included if a SELECT results in an empty set;
      defaults to ``Unable to find a matching {model}`` and supports both
      substitutions that `exception_message` does

    :returns: a `decorator`_ closure, which will be called with the function to
      be `decorated`_ as its argument.  This is a regular callable decorator.

    :rtype: callable which wraps and returns a function

    The decorator sets up some global symbols for use inside the DB access function:

      :session: a :class:`~sqlalchemy.orm.Session` instance created in a
        context containing the execution of the wrapped coroutine, suitable for
        performing database interactions, and which will be automatically
        closed after the coroutine completes

      :model_name: a string containing the lowercase name of the model

    .. _decorator: https://docs.python.org/3/glossary.html#term-decorator
    .. _decorated: https://peps.python.org/pep-0318/
    .. _coroutine: https://docs.python.org/3/library/asyncio-task.html#coroutines

    """

    mname = model_name(cls)
    Session = sessionmaker(engine)

    def db_func_wrapper(db_func):
        exc = exception_message.format(
            route_name=db_func.__name__, model=mname
        )
        empty = empty_set_message.format(
            route_name=db_func.__name__, model=mname
        )

        @wraps(db_func)
        def wrapped_interaction(*args, **kwargs):
            with Session() as session:
                db_func.__globals__["session"] = session
                db_func.__globals__["model_name"] = mname

                try:
                    result = db_func(*args, **kwargs)
                    if result:
                        return result
                except HTTPException as e:
                    raise e
                except Exception:
                    logger.exception(exc + f"({args!r},{kwargs!r})")
            raise HTTPException(status_code=404, detail=empty)

        wrapped_interaction.__doc__ = db_func.__doc__
        return wrapped_interaction  # a regular function

    return db_func_wrapper  # a regular function


def db_interaction(  # a decorator with parameters
    *,
    cls,
    engine=sql_engine,
    exception_message: str = ("{route_name}:{model}"),
    empty_set_message: str = ("Unable to find a matching {model}"),
):
    """Decorator for database interactions

    :param ~chapps.models.CHAPPSModel cls: the data model class

    :param ~sqlalchemy.engine.Engine engine: an :mod:`SQLAlchemy` engine, which
      defaults to the package-wide one declared in
      :mod:`~chapps.dbsession`

    :param str exception_message: a message to include if any untrapped
      exception occurs; defaults to ``{route_name}:{model}``.  Only those two
      symbols are available for expansion.  All arguments are appended.

    :param str empty_set_message: included if a SELECT results in an empty set;
      defaults to ``Unable to find a matching {model}`` and supports both
      substitutions that `exception_message` does

    :returns: a `decorator`_ closure, which will be called with the function to
      be `decorated`_ as its argument.  In this case, the function is expected
      to be a coroutine which is being manufactured for use in the API, and so
      the decorator closure returned by this routine defines a `coroutine`_ to
      wrap and await its argument, which is ultimately returned and used as the
      API route.

    :rtype: callable which wraps and returns a coroutine

    The decorator sets up some global symbols for use inside the API route
    coroutines:

      :session: a :class:`~sqlalchemy.orm.Session` instance created in a
        context containing the execution of the wrapped coroutine, suitable for
        performing database interactions, and which will be automatically
        closed after the coroutine completes

      :model_name: a string containing the lowercase name of the model

    .. _decorator: https://docs.python.org/3/glossary.html#term-decorator
    .. _decorated: https://peps.python.org/pep-0318/
    .. _coroutine: https://docs.python.org/3/library/asyncio-task.html#coroutines

    """

    mname = model_name(cls)
    Session = sessionmaker(engine)

    def interaction_wrapper(rt_coro):
        exc = exception_message.format(
            route_name=rt_coro.__name__, model=mname
        )
        empty = empty_set_message.format(
            route_name=rt_coro.__name__, model=mname
        )

        @wraps(rt_coro)
        async def wrapped_interaction(*args, **kwargs):
            with Session() as session:
                rt_coro.__globals__["session"] = session
                rt_coro.__globals__["model_name"] = mname

                try:
                    result = await rt_coro(*args, **kwargs)
                    if result:
                        return result
                except HTTPException as e:
                    raise e
                except Exception:
                    logger.exception(exc + f"({args!r},{kwargs!r})")
            raise HTTPException(status_code=404, detail=empty)

        wrapped_interaction.__doc__ = rt_coro.__doc__
        return wrapped_interaction  # a coroutine

    return interaction_wrapper  # a regular function


def get_item_by_id(
    cls,
    *,
    response_model,
    engine=sql_engine,
    assoc: Optional[List[JoinAssoc]] = None,
):
    """Build a route coroutine to get an item by ID

    :param ~chapps.models.CHAPPSModel cls: the main data model for the request

    :param ~chapps.models.CHAPPSResponse response_model: the response
      model

    :param ~sqlalchemy.engine.Engine engine: defaults to
      :const:`~chapps.dbsession.sql_engine`

    :param List[~chapps.rest.dbmodels.JoinAssoc] assoc: if included, these
      associations will be included as optional keys in the response

    At present there is no provision for dealing with extremely long
    association lists.  Even if there were 500 elements, the response would not
    be extremely large.

    .. note::

      An alternate closure factory for creating routes which
      specifically list associations does provide pagination, etc.  See
      :func:`~.list_associated`

    The factory produces a coroutine decorated with the
    :func:`~.db_interaction` decorator, as do all the route factories.  Its
    signature is:

      .. code:: python

        async def get_i(item_id: int) -> response_model

    The factory sets the final closure's name and doc metadata properly to
    ensure that the automatic documentation is coherent and accurate.  All the
    route factories do this to a greater or lesser extent.

    .. todo::

      provide option for API user to suppress each association
      perhaps something like `no_list_domains` as part of the query params

    """
    mname = model_name(cls)

    @db_interaction(cls=cls, engine=engine)
    async def get_i(item_id: int):
        stmt = cls.select_by_id(item_id)
        item = session.scalar(stmt)
        if item:
            if assoc:
                extra_args = {
                    a.assoc_name: a.assoc_model.wrap(
                        getattr(item, a.assoc_name)
                    )
                    for a in assoc
                }
                return response_model.send(cls.wrap(item), **extra_args)
            else:
                return response_model.send(cls.wrap(item))

    get_i.__name__ = f"get_{model_name(cls)}"
    get_i.__doc__ = (
        f"Retrieve **{mname}** records by ID, "
        "along with all (up to 1000) associated records."
    )
    return get_i


def list_items(cls, *, response_model, engine=sql_engine):
    """Build a route coroutine to list items

    :param ~chapps.models.CHAPPSModel cls: the main data model for the
      request

    :param ~chapps.models.CHAPPSResponse response_model: the response
      model

    :param ~sqlalchemy.engine.Engine engine: defaults to
      :const:`~chapps.dbsession.sql_engine`

    The returned closure expects to receive the query parameters as a dict,
    since that is what the dependency will return.  Its signature is

      .. code:: python

        async def list_i(qparams: dict = Depends(list_query_params))

    The closure's name and document metadata are updated to ensure coherence
    and accuracy of the automatic API documentation.

    For an example of using this factory, see :ref:`Listing Domains
    <listing-domains>`.

    """
    mname = model_name(cls)

    @db_interaction(cls=cls, engine=engine)
    async def list_i(qparams: dict = Depends(list_query_params)):
        stmt = cls.windowed_list(**qparams)
        items = cls.wrap(session.scalars(stmt))
        if items:
            return response_model.send(items)

    list_i.__name__ = f"list_{model_name(cls)}"
    list_i.__doc__ = f"""
        List **{mname}** records.<br/>
        Pass a substring to match as `q`.<br/>
        Paginate by providing `skip` and `limit`.
        """
    return list_i


def list_associated(
    cls: CHAPPSModel,
    *,
    assoc: JoinAssoc,
    response_model: CHAPPSResponse,
    engine=sql_engine,
):
    """Build a route to list associated items with pagination

    :param cls: the main data model

    :param assoc: the association to list

    :param response_model: the response model

    :param sqlalchemy.engine.Engine engine: defaults to
      :const:`~chapps.dbsession.sql_engine`

    The returned coroutine will paginate a list of the associated objects,
    given the ID of a main (source) object to use to select associations.  The
    `qparams` parameter is a bundle of standard listing query parameters
    defined by :func:`.list_query_params` via the :class:`fastapi.Depends`
    mechanism.

      .. code:: python

        async def assoc_list(item_id: int, qparams: dict) -> response_model

    It returns in the `response` key of its output a list of
    the associated object, goverened by the search and window parameters in
    `qparams`.

    """
    mname = model_name(cls)
    fname = f"{mname}_list_{assoc.assoc_name}"

    @db_interaction(cls=cls, engine=engine)
    async def assoc_list(
        item_id: int, qparams: dict = Depends(list_query_params)
    ):
        stmt = assoc.assoc_model.windowed_list_by_ids(
            subquery=assoc.select_ids_by_source_id(item_id), **qparams
        )
        assocs = assoc.assoc_model.wrap(session.scalars(stmt))
        if assocs:
            return response_model.send(assocs)

    assoc_list.__name__ = fname
    assoc_list.__doc__ = (
        f"List **{assoc.assoc_name}** associated with a particular **{mname}**,"
        " identified by ID.<br/>Supply standard query parameters for matching and"
        " pagination."
    )
    return assoc_list


def delete_item(cls, *, response_model=DeleteResp, engine=sql_engine):
    """Build a route coroutine to delete an item by ID

    :param ~chapps.models.CHAPPSModel cls: the data model to manage

    :param ~chapps.models.CHAPPSResponse response_model: defaults to :class:`~chapps.models.DeleteResp`

    :param ~sqlalchemy.engine.Engine engine: defaults to :const:`~chapps.dbsession.sql_engine`

    The returned coroutine accepts a list of record IDs for the specified
    object type and delete them.  Its signature is:

      .. code:: python

        async def delete_i(item_ids: List[int]) -> DeleteResp

    """
    mname = model_name(cls)

    @db_interaction(cls=cls, engine=engine)
    async def delete_i(item_ids: List[int]):
        try:
            session.execute(cls.remove_by_id(item_ids))
            session.commit()
        except IntegrityError:
            logger.exception("trying to delete {mname} with {args!r}")
            raise HTTPException(
                status_code=status.HTTP_409_CONFLICT,
                detail="Database integrity conflict.",
            )
        return response_model.send()

    delete_i.__name__ = f"delete_{mname}"
    delete_i.__doc__ = f"""
        Accepts a list of **{mname}** object IDs.
        Deletes the records with those IDs."""
    return delete_i


def adjust_associations(
    cls,
    *,
    assoc: List[JoinAssoc],
    assoc_op: AssocOperation,
    params: dict = None,
    response_model=TextResp,
    engine=sql_engine,
):
    """Build a route to add or subtract association lists, or set unitary ones

    :param ~chapps.models.CHAPPSModel cls: a data model class

    :param List[~chapps.rest.dbmodels.JoinAssoc] assoc: list of associations to
      operate on

    :param ~chapps.models.AssocOperation assoc_op: operation to perform on
      the association

    :param ~chapps.models.CHAPPSResponse response_model: the response
      model to send

    :param ~sqlalchemy.engine.Engine engine: defaults to
      :const:`~chapps.dbsession.sql_engine`

    The returned coroutine provides logic for a route which adds or subtracts
    elements to or from those already associated with the main object.  Its
    exact signature is dependent on what associations are listed.  After
    `item_id`, which is an ID to use to look up the main object, it will expect
    further arguments named as the association (`assoc_name`) which are of the
    specified type (`assoc_type`).

    If only one association is adjusted by the route, there will be just the
    one list (or scalar) as a body argument, which doesn't get a label, making
    the API call very easy to format and looking very clean in the API docs.
    If more than one are specified, :mod:`FastAPI` will expect a JSON object in
    the body with keys named as the ID columns and values which are lists of
    IDs.

    It all seems quite complicated when stated this way, but when viewed in the
    API documentation, it makes much more sense.

    For an example of using this factory, see :ref:`Handling Associations
    <handling-associations>`

    """

    mname = model_name(cls)
    assoc_s = "_".join([a.assoc_name for a in assoc])
    fname = f"{mname}_{assoc_op}_{assoc_s}"
    params = params or dict(item_id=int)
    assoc_params = {a.assoc_name: a.assoc_type for a in assoc}

    @db_interaction(cls=cls, engine=engine)
    async def assoc_op_i(*pargs, **args):
        # item_id will be used for the source object, and assoc_names will
        # be a list of associated ids to either remove or add associations
        # for, ignoring integrity errors arising from attempting to insert
        # duplicate associations; non-existent associations should not cause
        # errors when a query attempts to delete them.
        extras = {
            a.assoc_name: (a, args.pop(a.assoc_name))
            for a in assoc
            if a.assoc_name in args
        }
        item_id = args["item_id"]
        if extras:
            for assoc_name, (assc, vals) in extras.items():
                if not vals:
                    continue
                if assoc_op == AssocOperation.add:
                    stmt = assc.insert_assoc(item_id, vals)
                elif assoc_op == AssocOperation.replace:
                    stmt = assc.update_assoc(item_id, vals)
                else:
                    stmt = assc.delete_assoc(item_id, vals)
                try:
                    session.execute(stmt)
                except IntegrityError:
                    pass  # ignoring as stated above
            session.commit()
            return response_model.send("Updated.")
        return response_model.send("Empty request.")

    # this approach may seem laborious, but it supports multicolumn prikeys
    routeparams = [
        inspect.Parameter(
            name=param,
            kind=inspect.Parameter.POSITIONAL_OR_KEYWORD,
            annotation=type_,
        )
        for param, type_ in params.items()
    ]
    routeparams.extend(
        [
            inspect.Parameter(
                name=param,
                kind=inspect.Parameter.POSITIONAL_OR_KEYWORD,
                annotation=type_,
            )
            for param, type_ in assoc_params.items()
        ]
    )
    assoc_op_i.__signature__ = inspect.Signature(routeparams)
    assoc_op_i.__annotations__ = params
    assoc_op_i.__name__ = fname
    assoc_op_i.__doc__ = f"""
        {str(assoc_op).capitalize()}
        **{', '.join([a.assoc_name for a in assoc])}** objects
        {'to' if assoc_op==AssocOperation.add else 'from'} **{mname}**<br/>
        Accepts **{mname}** ID as `item_id`.
    """

    return assoc_op_i


def update_item(
    cls, *, response_model, assoc: List[JoinAssoc] = None, engine=sql_engine
):
    """Build a route to update items.

    :param ~chapps.models.CHAPPSModel cls: the main data model

    :param ~chapps.models.CHAPPSResponse response_model: the response
      model

    :param ~chapps.rest.dbmodels.JoinAssoc assoc: the association to list

    :param ~sqlalchemy.engine.Engine engine: defaults to
      :const:`~chapps.dbsession.sql_engine`

    The returned coroutine implements an API route for updating an item by ID,
    optionally including any associations included when the route coroutine is
    built.  If association data is provided to the route, it will completely
    replace any existing associations to that type of record with the new list
    of IDs.

    Its signature is determined by the contents of the
    :class:`~chapps.rest.dbmodels.JoinAssoc` passed to it.  The factory
    constructs :mod:`~inspect.Parameter` elements and uses them to create a
    correct :mod:`~inspect.Signature` for the new, decorated closure.  It also
    sets the `__doc__` and `__name__` metadata so that `FastAPI`_ will be able
    to find all the required data to create an API route with good
    documentation.

    For an example of how to use this factory, see :ref:`Updating Domains
    <updating-domains>`

    .. todo::

      in a generalized version of this for wider use in gluing `SQLAlchemy`_ to
      `FastAPI`_, it would need to allow arbitrary attributes of the model to
      be optional/required/defaulted.  This might easily be achieved through
      the use of an additional alternate Pydantic_ data model for updates,
      wherein those elements which ought to be optional may be marked as such.

    """
    mname = model_name(cls)
    fname = f"update_{mname}"
    params = {mname: cls}  # we are updating objects of this type

    @db_interaction(cls=cls, engine=engine)
    async def update_i(*pargs, **args):
        extras = {}
        assoc_ret = {}
        if assoc:
            extras = {
                a.assoc_name: (a, args.pop(a.assoc_name))
                for a in assoc
                if a.assoc_name in args
            }
        item_id = args[mname].id
        stmt = cls.update_by_id(args[mname])
        try:
            result = session.execute(stmt)
            if result.rowcount > 0:  # commit if changes were made
                session.commit()
            else:  # return None if there is no record to update -> 404
                return
        except Exception:
            logger.exception(f"{fname}: {args!r}")
            raise HTTPException(
                status_code=status.HTTP_500_INTERNAL_SERVER_ERROR
            )
        item = session.scalar(cls.select_by_id(item_id))
        if extras:
            for assoc_name, (assc, vals) in extras.items():
                if not vals:
                    continue
                try:
                    session.execute(
                        assc.delete().where(
                            getattr(assc.table.c, assc.source_id) == item_id
                        )
                    )
                    session.execute(assc.insert(), assc.values(item, vals))
                    session.commit()
                    assoc_ret[assoc_name] = assc.assoc_model.wrap(
                        getattr(item, assoc_name)
                    )
                except IntegrityError:
                    logger.exception(
                        f"{fname}: associating {item} with"
                        f" {assoc_name}s {vals!r}"
                    )
                    raise HTTPException(
                        status_code=status.HTTP_409_CONFLICT,
                        detail=(
                            "Unable to create requested association to"
                            f"{assc.assoc_model.__name__.lower()} entries."
                            "  Please check object ids and try again."
                        ),
                    )
        return response_model.send(cls.wrap(item), **assoc_ret)

    routeparams = [
        inspect.Parameter(
            name=param,
            kind=inspect.Parameter.POSITIONAL_OR_KEYWORD,
            default=Body(...),
            annotation=type_,
        )
        for param, type_ in params.items()
    ]
    if assoc:
        routeparams.extend(
            [
                inspect.Parameter(
                    name=a.assoc_name,
                    kind=inspect.Parameter.POSITIONAL_OR_KEYWORD,
                    default=Body(None),
                    annotation=a.assoc_type,
                )
                for a in assoc
            ]
        )
    update_i.__signature__ = inspect.Signature(routeparams)
    update_i.__annotations__ = params
    update_i.__name__ = fname
    update_i.__doc__ = f"""
        Update a **{mname}** record by ID.<br/>
        Only the **id** and the adjusted attribute(s) are required.<br/>
        Associations are not required, but if provided (by ID), will
        completely replace any existing association relationships
        of the same type.
        """
    return update_i


def create_item(
    cls,
    *,
    response_model,
    params=None,
    defaults=None,
    assoc=None,
    engine=sql_engine,
):
    """Build a route coroutine to create new item records

    :param ~chapps.models.CHAPPSModel cls: the main data model

    :param ~chapps.models.CHAPPSResponse response_model: the response
      model

    :param dict params: defaults to ``dict(name=str)``; specify to provide
      additional column names and types, and be sure to include `name`, as all
      models currently are expected to have a `name` column, which is not
      allowed to be null.

    :param ~.JoinAssoc assoc: the associations to attach,
      if any

    :param ~sqlalchemy.engine.Engine engine: defaults to
      :const:`~chapps.dbsession.sql_engine`

    The returned coroutine implements an API route for creating an item,
    setting all its elements (other than ID) to whatever values are provided.
    Currently all values must be provided.  If desired, associations may also
    be provided to the factory, and they will be accommodated by the coroutine.

    For an example invocation of this factory, see :ref:`Creating Users <creating-users>`

    """
    params = params or dict(name=str)
    defaults = defaults or dict()
    mname = model_name(cls)
    fname = f"create_{mname}"

    @db_interaction(cls=cls, engine=engine)
    async def create_i(*pargs, **args):
        extras = {}
        assoc_ret = {}
        if assoc:
            extras = {
                a.assoc_name: (a, args.pop(a.assoc_name))
                for a in assoc
                if a.assoc_name in args
            }
        item = cls.Meta.orm_model(**args)
        try:
            session.add(item)
            session.commit()
        except IntegrityError:
            raise HTTPException(
                status_code=status.HTTP_409_CONFLICT,
                detail=f"Unique key conflict creating {mname}.",
            )
        except OperationalError as e:
            raise HTTPException(
                status_code=status.HTTP_400_BAD_REQUEST, detail=str(e)
            )
        except Exception:
            logger.exception(f"{fname}: {args!r}")
            raise HTTPException(
                status_code=status.HTTP_500_INTERNAL_SERVER_ERROR
            )
        if extras:
            for assoc_name, (assc, vals) in extras.items():
                if not vals:  # no empty inserts
                    continue
                try:
                    session.execute(assc.insert(), assc.values(item, vals))
                    session.commit()
                    assoc_ret[assoc_name] = assc.assoc_model.wrap(
                        getattr(item, assoc_name)
                    )
                except IntegrityError:
                    raise HTTPException(
                        status_code=status.HTTP_409_CONFLICT,
                        detail=(
                            "Unable to create requested association to"
                            f"{assc.assoc_model.__name__.lower()} entries."
                            "  Please check object ids and try again."
                        ),
                    )

        return response_model.send(cls.wrap(item), **assoc_ret)

    routeparams = [  # assemble signature for FastAPI
        inspect.Parameter(
            name=param,
            kind=inspect.Parameter.POSITIONAL_OR_KEYWORD,
            default=Body(defaults.get(param, ...)),
            annotation=type_,
        )
        for param, type_ in params.items()
    ]
    if assoc:
        routeparams.extend(
            [
                inspect.Parameter(
                    name=a.assoc_name,
                    kind=inspect.Parameter.POSITIONAL_OR_KEYWORD,
                    default=Body(None),
                    annotation=a.assoc_type,
                )
                for a in assoc
            ]
        )
    defaulted_attrs = [a for a, d in defaults.items() if d != ...]
    if len(defaulted_attrs):
        dfl_attr_doc = (
            "The `name` attribute is required, along with any others"
            + " not described below as optional.<br/>"
            + "The following attributes are <b>optional</b>:<br/>"
            + f"<ul>{' '.join(['<li>`'+e+'`</li>' for e in defaulted_attrs])}"
            + "</ul>"
        )
    else:
        dfl_attr_doc = "All attributes are required."
    create_i.__signature__ = inspect.Signature(routeparams)
    create_i.__annotations__ = params
    create_i.__name__ = fname
    create_i.__doc__ = (
        f"Create a new **{mname}** record in the database.<br/>"
        + "The new object will be returned, including its ID.  "
        + "Raises descriptive errors on 409; checking the detail "
        + "of the error may aid in debugging.<br/>"
        + dfl_attr_doc
    )
    return create_i



"""
Logging
-------

This module's purpose is to ensure that CHAPPS modules
end up instantiating loggers which all share a common
formatter and handler.

It has been reworked a little bit since its inception,
and it may approach the issue from two directions, still.
It does attempt to be the first thing to get a handler,
and thereby be able to set the basic config.

But it also instantiates a logger at the top of the hierarchy,
one named simply `chapps`, so that all further CHAPPS loggers
(using the recommended pattern) will be under that parent logger
and inherit its config.

"""
import logging
from logging.handlers import SysLogHandler

DEFAULT_LEVEL = logging.DEBUG
"""Default minimum severity `DEBUG`"""

DEFAULT_FACILITY = SysLogHandler.LOG_LOCAL0
"""Default **syslog** facility `LOCAL0`"""


class LogSetup:  # pragma: no cover
    """Establish global log format and handler for CHAPPS

    For consistency, logs from different modules should all appear to come from
    CHAPPS, otherwise it may be quite a headache to make sense of them.  The
    class defines two different formatters, one for debugging and one for
    general use.  There is no fancy method in place for switching between them
    at present.

    The default formatter is very basic, simply prepending "CHAPPS:" and the
    severity level to the message.

    The default minimum severity at present is `DEBUG`.

    The default **syslog** facility is `LOCAL0`.

    It is possible to use **rsyslog** configuration to send different levels of
    the `LOCAL0` facility to different destinations, or to ignore `DEBUG`
    level messages entirely.  An example config is included in the project's
    ancillary materials.

    It had been a conscious decision not to include the :mod:`chapps.config`
    module in this module, in order to avoid circular dependencies when this
    module was included everywhere.  The rework has the config module loaded as
    part of package initialization, so it is no longer necessary to load it
    anywhere, since once it is loaded its job is done.

    .. todo::

      add configuration support for **syslog** facility and minimum severity.

    """

    debug_formatter = logging.Formatter(
        "CHAPPS:%(levelname)s %(filename)s@%(lineno)s: %(message)s"
    )
    """A formatter helpful for debugging."""

    maillog_formatter = logging.Formatter(
        "%(processName)s[%(process)d]: CHAPPS %(levelname)s %(message)s"
    )
    """The default formatter, anticipated to go to the mail log"""

    syslog_handler = SysLogHandler(
        facility=DEFAULT_FACILITY, address="/dev/log"
    )
    """A handler which uses `/dev/log` to send messages to **syslog**"""

    syslog_handler.setLevel(DEFAULT_LEVEL)
    syslog_handler.setFormatter(maillog_formatter)  # or debug_formatter

    def __init__(self):
        """Setup logging

        If there are not yet any handlers, this routine calls
        :py:func:`logging.basicConfig` to set up basic logging configuration.

        If there are already handlers, for instance due to running within
        :mod:`pytest`, then nothing happens.

        """
        if not logging.getLogger(None).hasHandlers():
            logging.basicConfig(handlers=[self.syslog_handler])


LogSetup()
logger = logging.getLogger("chapps")
"""Create a logger in order to configure the top of the hierarchy"""
# without this, the library may not emit logs from a script
logger.setLevel(DEFAULT_LEVEL)



#!/usr/bin/env python3
"""
Starts a service on 127.0.0.1:10025 which serves as a null content filter
"""
### requires the python-pidfile library from https://github.com/mosquito/python-pidfile
### requires aiosmtpd
import asyncio, pidfile, signal, functools
from smtplib import SMTP, SMTPRecipientsRefused
from aiosmtpd.smtp import SMTP as SMTPServer
from aiosmtpd.handlers import Proxy, Debugging
from aiosmtpd.controller import Controller
import logging
from logging.handlers import SysLogHandler

LISTEN_ADDR = "localhost"
LISTEN_PORT = 10025
TRANSMIT_ADDR = LISTEN_ADDR
TRANSMIT_PORT = 10026


logging.basicConfig(
    handlers=[SysLogHandler(facility=SysLogHandler.LOG_LOCAL4, address="/dev/log")]
)

logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)

SMTPServer_HANDLER = functools.partial(SMTPServer, Proxy(TRANSMIT_ADDR, TRANSMIT_PORT))


def signal_handler(sig, *args):
    if sig in {signal.SIGTERM, signal.SIGINT}:
        logger.debug(f"null-filter exiting on signal {sig}.")
        raise SystemExit


def install_asyncio_signal_handlers(loop):
    for signame in {"SIGTERM", "SIGINT"}:
        sig = getattr(signal, signame)
        loop.add_signal_handler(sig, functools.partial(signal_handler, sig))


# class NullFilterHandler:
#     async def handle_RCPT(self, server, session, envelope, address, rcpt_options):
#         """Handle recipient phase"""
#         envelope.rcpt_tos.append( address )
#         return "250 OK"

#     async def handle_DATA(self, server, session, envelope):
#         """Handle DATA phase"""
#         logger.debug(f"Message from {envelope.mail_from} to ")
#         try:
#             client = SMTP.sendmail( envelope.mail_from, envelope.rcpt_tos, envelope.content )
#             return '250 Message accepted for delivery'
#         except smtplib.SMTPResponseException as e:
#             logger.exception("Upstream Postfix did not like this message.")
#             return f"{e.smtp_code} {e.smtp_error}"
#         except smtplib.SMTPException:
#             logger.exception("Raised trying to send from {envelope.mail_from} to {','.join(envelope.rcpt_tos)}")
#             return "550 Requested action not taken"


async def main():
    """The grand shebang"""
    logger.debug("Starting null-filter...")
    try:
        with pidfile.PIDFile("/tmp/null_filter.pid"):
            logger.debug("null-filter started.")
            loop = asyncio.get_running_loop()
            install_asyncio_signal_handlers(loop)
            srv = await loop.create_server(
                SMTPServer_HANDLER, LISTEN_ADDR, LISTEN_PORT, start_serving=False
            )
            await srv.serve_forever()
    except TimeoutError:
        logger.exception("SMTPServer took too long to start.")
    except pidfile.AlreadyRunningError:
        logger.exception("null-filter is already running. Exiting.")
    except asyncio.exceptions.CancelledError:
        logger.debug("null-filter exiting on signal.")


if __name__ == "__main__":
    try:
        asyncio.run(main())
    except Exception:
        logger.exception("UNEXPECTED")



"""
**Quota** record management implemented by factories
----------------------------------------------------

This module defines the API router for **Quota** record manipulation,
and defines the :class:`~.JoinAssoc` which describes the relationship between
**Quota** and **User** tables.

Implementation of **Quota** routes is a little simpler than for other models
because some functionality is intentionally excluded.  It is expected that a
large number of users might share the same **Quota** record, therefore it is
not supported to retrieve the **User** records associated with a **Quota**
object.

"""

from fastapi import APIRouter, status
from chapps.models import Quota, QuotaResp, QuotasResp, DeleteResp
from chapps.rest.routers.common import (
    get_item_by_id,
    list_items,
    create_item,
    delete_item,
    update_item,
)
import logging
import chapps.logging

logger = logging.getLogger(__name__)
logger.setLevel(chapps.logging.DEFAULT_LEVEL)

api = APIRouter(
    prefix="/quotas",
    tags=["quotas"],
    responses={404: {"description": "Quota not found."}},
)
"""The **Quota** record management API router"""

api.get("/", response_model=QuotasResp)(
    list_items(Quota, response_model=QuotasResp)
)


api.get("/{item_id}", response_model=QuotaResp)(
    get_item_by_id(Quota, response_model=QuotaResp)
)

api.post(
    "/",
    status_code=201,
    response_model=QuotaResp,
    responses={status.HTTP_409_CONFLICT: {"description": "Unique key error."}},
)(
    create_item(
        Quota, response_model=QuotaResp, params=dict(name=str, quota=int)
    )
)

api.delete("/", response_model=DeleteResp)(delete_item(Quota))

api.put("/", response_model=QuotaResp)(
    update_item(Quota, response_model=QuotaResp)
)



"""
**User** record management implemented by factories
---------------------------------------------------

This module defines API routes for managing **User** records, and defines the :class:`JoinAssoc` data to describe the relationship between **User** records and other record types: **Domain**, **Email**, and **Quota**.

As such, this is the only router where the factories are used to create or update records with multiple associations.  It seems a good opportunity for another example.

"""

from typing import List, Optional
from fastapi import status, APIRouter
from chapps.models import (
    User,
    Quota,
    Domain,
    Email,
    UserResp,
    UsersResp,
    DomainsResp,
    EmailsResp,
    DeleteResp,
    IntResp,
    TextResp,
    AssocOperation,
    BulkQuotaResp,
    BulkDomainsResp,
    BulkEmailsResp,
    user_quota_assoc,
    user_domains_assoc,
    user_emails_assoc,
)
from chapps.rest.routers.common import (
    get_item_by_id,
    list_items,
    create_item,
    delete_item,
    update_item,
    adjust_associations,
    list_associated,
    load_models_with_assoc,
)
import logging
import chapps.logging

logger = logging.getLogger(__name__)

api = APIRouter(
    prefix="/users",
    tags=["users"],
    responses={404: {"description": "User not found."}},
)
"""The API router for **User** record maintenance

This router is once again full of calls to the factories in the :mod:`~.common`
module.  The **User** model is the most connected to other things, however, and
so seems like a good spot for examples related to associations.

.. _creating-users:
.. rubric:: Creating Users

When creating a **User** record, it seems likely that the caller might like to
automatically associate the new **User** with an existing **Quota** record, and
at least one **Domain** record, perhaps even an **Email** record.  The factory
will provide a coroutine which will optionally accept ID lists for these
associations.  That is to say, the coroutine will treat them as optional
arguments and do nothing if they are not provided.

All of the logic and magic behind making this go smoothly is hidden within the
:class:`.JoinAssoc` class.  We simply provide a list of these associations to
the factory and it handles the rest:

.. code:: python

    api.post(
        "/",
        status_code=201,
        response_model=UserResp,
        responses={status.HTTP_409_CONFLICT: {"description": "Unique key error."}},
    )(create_item(User, response_model=UserResp, assoc=user_join_assoc))

In the above example, the `FastAPI`_ code for specifying details about the POST
route takes up more space than the factory call to obtain the actual **User**
creation coroutine.

The definition of :const:`.user_join_assoc` may be found below.  It is a list
containing references to all three :class:`~.JoinAssoc` instances, relating to
a **Quota** and lists of **Domain** and **Email** records.

.. _handling-associations:
.. rubric:: Handling Associations

Sometimes there is a need to remove a specific association from a list, or add
one or a handful.  It would be helpful if it were not necessary to obtain or
manufacture a list of IDs in order to use a replacement-type edit such as the
basic model update route.  The **User** model has a number of different
associations to manage, so here is an example of adding domains:

.. code:: python

    api.put("/{item_id}/domains/", response_model=TextResp)(
        adjust_associations(
            User, assoc=[user_domains_assoc], assoc_op=AssocOperation.add
        )
    )

I chose to use PUT because it is associated with partial updates.  Within the
API router wrapper, we use a call to the :func:`~.adjust_associations` route
factory, which returns a coroutine which will take a **User** ID and a list of
**Domain** IDs as arguments.  When invoked via the API, that coroutine will
ensure that all the existing **Domain** records listed are associated to the
**User**.  :exc:`~sqlalchemy.exc.IntegrityError` is ignored during the process,
so any attempts to add an existing association or to add a nonexistent
**Domain** will not raise errors -- all existing **Domain** records identified
by ID will be associated to the **User**, and other associations to that
**User** will be preserved.

"""

user_join_assoc = [user_quota_assoc, user_domains_assoc, user_emails_assoc]

load_users_with_quota = load_models_with_assoc(User, assoc=user_quota_assoc)
load_users_with_domains = load_models_with_assoc(
    User, assoc=user_domains_assoc
)
load_users_with_emails = load_models_with_assoc(User, assoc=user_emails_assoc)

api.post(
    "/",
    status_code=201,
    response_model=UserResp,
    responses={status.HTTP_409_CONFLICT: {"description": "Unique key error."}},
)(create_item(User, response_model=UserResp, assoc=user_join_assoc))


api.delete("/", response_model=DeleteResp)(delete_item(User))


api.get("/", response_model=UsersResp)(
    list_items(User, response_model=UsersResp)
)


@api.get("/quotas/", response_model=BulkQuotaResp)
async def map_usernames_to_quota_ids(user_ids: List[int]):
    """Map **User** identfiers onto **Quota** ids

    If a display requires a large matrix of users with their quota settings,
    this routine may be helpful.  The **Quota** records may be fetched before
    or after, just once for each kind of quota, and then cross-referenced much
    more efficiently than requesting each separately.

    The `response` contains a list of JSON objects (hashes or dictionaries),
    with the keys `user_name` and `quota_id`.  Only existing users are
    returned, possibly with `quota_id` set to `None` if the user has no quota
    policy assigned.  They are sorted by the user's ID value.

    """
    users_with_quotas = load_users_with_quota(user_ids)
    if not users_with_quotas:
        return BulkQuotaResp.send([], ["No listed user IDs existed."])
    uqm = [
        {"user_name": u.name, "quota_id": u.quota.id if u.quota else None}
        for u in users_with_quotas
        if u
    ]
    return BulkQuotaResp.send(uqm)


@api.get("/domains/", response_model=BulkDomainsResp)
async def map_usernames_to_domain_ids(user_ids: List[int]):
    """Map **User** identfiers onto **Domain** id lists

    If a display requires a large matrix of users with their domain authorizations,
    this routine may be helpful.  The **Domain** records may be fetched before
    or after, just once for each domain, and then cross-referenced much
    more efficiently than requesting each separately.

    The `response` contains a list of JSON objects (hashes or dictionaries),
    with the keys `user_name` and `domain_ids`.  Only existing users are
    returned, possibly with `domain_ids` set to `None` if the user has no
    domain authorizations.  They are sorted by the user's ID value.

    """
    users_with_domains = load_users_with_domains(user_ids)
    if not users_with_domains:
        return BulkDomainsResp.send([], ["No listed user IDs existed."])
    udm = [
        {"user_name": u.name, "domain_ids": [d.id for d in u.domains]}
        for u in users_with_domains
        if u
    ]
    return BulkDomainsResp.send(udm)


@api.get("/emails/", response_model=BulkEmailsResp)
async def map_usernames_to_email_ids(user_ids: List[int]):
    """Map **User** identfiers onto **Email** id lists

    If a display requires a large matrix of users with their email authorizations,
    this routine may be helpful.  The **Email** records may be fetched before
    or after, just once for each email, and then cross-referenced much
    more efficiently than requesting each separately.

    The `response` contains a list of JSON objects (hashes or dictionaries),
    with the keys `user_name` and `email_ids`.  Only existing users are
    returned, possibly with `email_ids` set to `None` if the user has no
    email authorizations.  They are sorted by the user's ID value.

    """
    users_with_emails = load_users_with_emails(user_ids)
    if not users_with_emails:
        return BulkEmailsResp.send([], ["No listed user IDs existed."])
    uem = [
        {"user_name": u.name, "email_ids": [d.id for d in u.emails]}
        for u in users_with_emails
        if u
    ]
    return BulkEmailsResp.send(uem)


api.get("/{item_id}", response_model=UserResp)(
    get_item_by_id(User, response_model=UserResp, assoc=user_join_assoc)
)

api.get("/{item_id}/domains/", response_model=DomainsResp)(
    list_associated(User, assoc=user_domains_assoc, response_model=DomainsResp)
)

api.get("/{item_id}/emails/", response_model=EmailsResp)(
    list_associated(User, assoc=user_emails_assoc, response_model=EmailsResp)
)

api.put("/", response_model=UserResp)(
    update_item(User, response_model=UserResp, assoc=user_join_assoc)
)

api.put("/{item_id}/domains/", response_model=TextResp)(
    adjust_associations(
        User, assoc=[user_domains_assoc], assoc_op=AssocOperation.add
    )
)

api.put("/{item_id}/emails/", response_model=TextResp)(
    adjust_associations(
        User, assoc=[user_emails_assoc], assoc_op=AssocOperation.add
    )
)

api.delete("/{item_id}/emails/")(
    adjust_associations(
        User, assoc=[user_emails_assoc], assoc_op=AssocOperation.subtract
    )
)


api.delete("/{item_id}/domains/", response_model=TextResp)(
    adjust_associations(
        User, assoc=[user_domains_assoc], assoc_op=AssocOperation.subtract
    )
)

# note that the correct name of the quota parameter is necessary here
api.put("/{item_id}/quota/{quota}")(
    adjust_associations(
        User, assoc=[user_quota_assoc], assoc_op=AssocOperation.replace
    )
)

# commenting out to get a clean release without these non-working routes
#
# we will provide these routes in a future release
# along with routes to count a user's domain authorizations
# and paginate the list of those authorizations


# @api.get("/count/", response_model=IntResp)
# async def count_all_users():
#     return await count_users("%")


# @api.get("/count/{pattern}", response_model=IntResp)
# async def count_users(pattern: str):
#     cur = pca.conn.cursor()
#     sanitized_pattern = pattern
#     query = f"SELECT COUNT( * ) FROM users WHERE name LIKE ?"
#     logger.debug(
#         f"Attempting to count users like {pattern} with query {query}"
#     )
#     cur.execute(query, (f"%{pattern}%",))
#     results = cur.fetchone()[0]
#     cur.close()
#     return IntResp.send(results)



"""v0.4.12 DB setup

Revision ID: ce81767cec26
Revises:
Create Date: 2022-06-17 14:10:18.442341

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = "ce81767cec26"
down_revision = None
branch_labels = None
depends_on = None


def upgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table(
        "domains",
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column("name", sa.String(length=64), nullable=False),
        sa.PrimaryKeyConstraint("id", name=op.f("pk_domains")),
    )
    op.create_index(op.f("ix_domains_name"), "domains", ["name"], unique=True)
    op.create_table(
        "emails",
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column("name", sa.String(length=128), nullable=False),
        sa.PrimaryKeyConstraint("id", name=op.f("pk_emails")),
    )
    op.create_index(op.f("ix_emails_name"), "emails", ["name"], unique=True)
    op.create_table(
        "quotas",
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column("name", sa.String(length=32), nullable=False),
        sa.Column("quota", sa.Integer(), nullable=False),
        sa.PrimaryKeyConstraint("id", name=op.f("pk_quotas")),
    )
    op.create_index(op.f("ix_quotas_name"), "quotas", ["name"], unique=True)
    op.create_index(op.f("ix_quotas_quota"), "quotas", ["quota"], unique=True)
    op.create_table(
        "users",
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column("name", sa.String(length=128), nullable=False),
        sa.PrimaryKeyConstraint("id", name=op.f("pk_users")),
    )
    op.create_index(op.f("ix_users_name"), "users", ["name"], unique=True)
    op.create_table(
        "domain_user",
        sa.Column("user_id", sa.Integer(), nullable=False),
        sa.Column("domain_id", sa.Integer(), nullable=False),
        sa.ForeignKeyConstraint(
            ["domain_id"],
            ["domains.id"],
            name=op.f("fk_domain_user_domain_id_domains"),
            onupdate="CASCADE",
            ondelete="CASCADE",
        ),
        sa.ForeignKeyConstraint(
            ["user_id"],
            ["users.id"],
            name=op.f("fk_domain_user_user_id_users"),
            onupdate="RESTRICT",
            ondelete="CASCADE",
        ),
        sa.PrimaryKeyConstraint(
            "user_id", "domain_id", name=op.f("pk_domain_user")
        ),
    )
    op.create_table(
        "email_user",
        sa.Column("user_id", sa.Integer(), nullable=False),
        sa.Column("email_id", sa.Integer(), nullable=False),
        sa.ForeignKeyConstraint(
            ["email_id"],
            ["emails.id"],
            name=op.f("fk_email_user_email_id_emails"),
            onupdate="CASCADE",
            ondelete="CASCADE",
        ),
        sa.ForeignKeyConstraint(
            ["user_id"],
            ["users.id"],
            name=op.f("fk_email_user_user_id_users"),
            onupdate="RESTRICT",
            ondelete="CASCADE",
        ),
        sa.PrimaryKeyConstraint(
            "user_id", "email_id", name=op.f("pk_email_user")
        ),
    )
    op.create_table(
        "quota_user",
        sa.Column("user_id", sa.Integer(), nullable=False),
        sa.Column("quota_id", sa.Integer(), nullable=False),
        sa.ForeignKeyConstraint(
            ["quota_id"],
            ["quotas.id"],
            name=op.f("fk_quota_user_quota_id_quotas"),
            onupdate="CASCADE",
            ondelete="CASCADE",
        ),
        sa.ForeignKeyConstraint(
            ["user_id"],
            ["users.id"],
            name=op.f("fk_quota_user_user_id_users"),
            onupdate="RESTRICT",
            ondelete="CASCADE",
        ),
        sa.PrimaryKeyConstraint("user_id", name=op.f("pk_quota_user")),
    )
    # ### end Alembic commands ###


def downgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_table("quota_user")
    op.drop_table("email_user")
    op.drop_table("domain_user")
    op.drop_index(op.f("ix_users_name"), table_name="users")
    op.drop_table("users")
    op.drop_index(op.f("ix_quotas_quota"), table_name="quotas")
    op.drop_index(op.f("ix_quotas_name"), table_name="quotas")
    op.drop_table("quotas")
    op.drop_index(op.f("ix_emails_name"), table_name="emails")
    op.drop_table("emails")
    op.drop_index(op.f("ix_domains_name"), table_name="domains")
    op.drop_table("domains")
    # ### end Alembic commands ###



"""CHAPPS REST API implementation

Originally, I had hoped to isolate all the API-related stuff
in this package, but as I should have known, references to the
data models crept back into the main codebase even before
the API was finished.

The :mod:`~.api` module serves to supply some top-level documentation strings related
to the groups of API routes, and to include all the various API routers into the
final :class:`~fastapi.FastAPI` object.

.. todo::

  some of the 'API' code, mainly the data models and database access logic,
  needs to migrate into the main package.  Since access to the database is
  already isolated into the adapter classes, this should not be too tough to
  do.

"""
__all__ = ["api", "routers"]



"""
Utility classes
---------------

Since CHAPPS mainly deals with policy requests coming from Postfix,
there is a utility object for representing them in a way which makes
the code easier to read while also providing access optimizations.

There is another utility class for providing the configuration
data via an object which presents dictionary keys as attributes.

In order to create different defaults and access documentation files,
an object is provided which detects whether the library is running
within a virtual environment, and serves as a source of local paths
to package resources.

.. todo::

  add Postfix command class, to store action output along with
  status information.

"""
from collections.abc import Mapping
import re
import logging
import sys
import hashlib
from pathlib import Path
from typing import Any, Dict, List, Optional, Union
from chapps.signals import TooManyAtsException, NotAnEmailAddressException

logger = logging.getLogger(__name__)


def hash_password(password: str, encoding: str = "utf-8") -> str:
    return hashlib.sha256(password.encode(encoding)).hexdigest()


class VenvDetector:
    """Detect use of a virtual environment and calculate local paths

    The detector encapsulates the job of determining whether a virtual
    environment is activated, and assists in composing some important paths in
    order to locate certain files the application needs.

    One of those files is Markdown source which is imported into the live API
    documentation.  Another is the config file.

    Instance attributes:

      :datapath: :class:`~pathlib.Path` pointing at location where data files
        are installed

      :ve: :obj:`bool` indicating whether a virtual environment is active

      :docpath: :class:`~pathlib.Path` pointing at the location of the Markdown

      :confpath: :class:`~pathlib.Path` pointing at the config file

      :venvpath: :class:`~pathlib.Path` to the root of the active virtual
        environment, or None if none is active

    """

    def __init__(self, *, datapath: Optional[Union[str, Path]] = None):
        """Detect virtual environments and provide local package paths

        :param datapath: optional override to point
          at the data installation path of the package (or a surrogate).  The
          API uses this value to load the API readme into the live docs.

        """
        if datapath:
            self.datapath = Path(datapath)
        elif self.ve:
            self.datapath = Path(sys.prefix)
        else:
            self.datapath = Path("/usr/local")

    # find the base prefix; hopefully pyenv-compatible
    def get_base_prefix_compat(self) -> str:
        """Return the non-virtual base prefix

        Sometimes called `sys.real_prefix`, so we check for both.

        :returns: the base path prefix
        :rtype: str

        """
        return (
            getattr(sys, "base_prefix", None)
            or getattr(sys, "real_prefix", None)
            or sys.prefix
        )

    # return whether we are in a venv
    def in_virtualenv(self) -> bool:
        """Compare prefixes to determine if a virtual environment is active.

        :returns: true if a virtual environment is active, otherwise False

        :rtype: bool

        """
        return self.get_base_prefix_compat() != sys.prefix

    # return whether a Sphinx build launched the library
    def sphinx_build(self) -> bool:
        """Determine whether invoked by Sphinx

        :returns: `True` if Sphinx invoked the library

        """
        try:
            if __sphinx_build__:
                return True
        except NameError:
            return False

    @property
    def ve(self) -> bool:
        """Property which memoizes :meth:`~.in_virtualenv`"""
        if "_ve" not in vars(self):
            self._ve = self.in_virtualenv()
        return self._ve

    @property
    def sb(self) -> bool:
        """Property which memoizes :meth:`~.sphinx_build`"""
        if "_sb" not in vars(self):
            self._sb = self.sphinx_build()
        return self._sb

    @property
    def docpath(self) -> Path:
        """Memoizes the documentation location

        :returns: a :class:`~pathlib.Path` pointing at the Markdown files'
          location

        :rtype: pathlib.Path

        """
        if "_docpath" not in vars(self):
            self._docpath = self.datapath / "chapps"
        return self._docpath

    @property
    def confpath(self) -> Path:
        """Memoizes the config file's full path

        :returns: a :class:`~pathlib.Path` pointing at the config file

        :rtype: pathlib.Path


        """
        if "_confpath" not in vars(self):
            try:
                self._confpath = self.venvpath / "etc" / "chapps.ini"
            except TypeError:
                self._confpath = Path("/") / "etc" / "chapps" / "chapps.ini"
        return self._confpath

    @property
    def venvpath(self) -> Optional[Path]:
        """The virtual environment root, if any

        :returns: None or the value of :const:`sys.prefix` as a :class:`Path`

        :rtype: Optional[pathlib.Path]

        If no virtual environment is active, then `None` is returned,
        otherwise a :class:`Path` instance is returned, containing the path to
        the virtual environment.  This hasn't been tested with all types of
        virtual environment.

        """
        if self.ve:
            return Path(sys.prefix)


class AttrDict(Mapping):
    """Attribute Dictionary

    This simple class allows accessing the keys of a hash as attributes on an
    object.  As a useful side effect it also casts floats, integers and
    booleans in advance.

    This object is used in :class:`~chapps.config.CHAPPSConfig` for holding the configuration data.

    .. note::

      The purpose of this class is to map all the keys of a fairly small
      :obj:`dict` as attributes of the instance onto their values in the source
      dict.  This class does not perform the same sort of lazy-loading as the
      :class:`~.PostfixPolicyRequest` class below; it pre-maps all the elements
      in the source :obj:`dict`.  So be careful about passing large
      dictionaries to it.

    .. admonition:: Subclassing

      Given the stated purpose of the class, all *internal instance attributes*,
      i.e. ones not associated to a key-value pair in the source object, should
      begin with  `_` (an underscore).

    """

    boolean_pattern = re.compile("^[Tt]rue|[Ff]alse$")
    """A regex to detect text-string boolean values"""

    def __init__(
        self, data: Dict[str, Any] = None, **kwargs: Optional[Dict[str, Any]]
    ):
        r"""Populate an instance with attributes

        :param data: a :obj:`dict` mapping strings (attribute names) onto arbitrary values

        :param kwargs: arbitrary keyword arguments

        If, and only if, `data` is not provided, then the keyword arguments will be used in place of data provided as a :obj:`dict`.

        .. todo::

          add any `kwargs` to an existing `data` :obj:`dict`

        Henceforth whatever is rounded up to use shall be referred to as the `data`.

        The initialization routine creates an attribute on the instance for
        each key in the `data`, and then attempts to cast the value:

            1. to an :obj:`int`.

            2. If a :exc:`TypeError` is encountered, the unadulterated value is used.

            3. If only :exc:`ValueError` is raised, then it is casted to
               :obj:`float`

            4. if that causes another :exc:`ValueError` then it is matched
               against the :const:`.boolean_pattern` to see whether it matches,
               which is to say, whether it is a string containing "true" or
               "false".

            5. If so, a simple check is conducted to determine whether the
               match was four characters long: `True`.

            6. If it does not test positive for truth, it is considered to be
               `False`.

            7. But if it wasn't a match for the :const:`.boolean_pattern` at
               all, then its original value is preserved.

        Generally, instances of this class are used to present a particular
        module, such as a policy manager, with its configuration in a form
        which can be dereferenced with dot notation.  As such, values which
        cannot be casted to some other type are almost always left in their
        original form as `str`\ ings, because the `AttrDict` is being
        initialized with a sub-block of a :class:`~configparser.ConfigParser`
        as the source object, and its values will all be strings.

        """
        if not data:
            data = kwargs
        for k, v in data.items():
            if k[0:2] != "__":
                val = v
                try:
                    val = int(v)
                except ValueError:
                    try:
                        val = float(v)
                    except ValueError:
                        m = self.boolean_pattern.match(v)
                        if m:
                            val = (
                                m.span(0)[1] == 4
                            )  # if the match is 4 chars long, it is True
                except TypeError:
                    pass
                setattr(self, k, val)

    def __iter__(self):
        return self.keys()

    def __len__(self) -> int:
        return len(self.keys())

    def __getitem__(self, item) -> Optional[str]:
        return getattr(self, item)

    def __contains__(self, item):
        return item in self.__dict__

    def keys(self):
        return self.__dict__.keys()


class PostfixPolicyRequest(Mapping):
    """Lazy-loading Policy Request Mapping Interface

    An implementation of :class:`~collections.abc.Mapping` which by default
    only processes and caches values from the data payload when they are
    accessed, to avoid a bunch of useless parsing.  Instances may be
    dereferenced like hashes, but the keys are also attributes on the instance,
    so they can be accessed without brackets and quotation marks.

    Once parsed, results are memoized.

    For example, a payload might look a bit like this, when it is first
    received from Postfix and turned into an array of one string per line:

    .. code:: python

      payload = [
        "request=smtpd_access_policy",
        "protocol_state=RCPT",
        "protocol_name=SMTP",
        "helo_name=helo.chapps.io",
        "queue_id=8045F2AB23",
        "sender=unauth@easydns.com",
        "recipient=bar@foo.tld",
        "recipient_count=0",
        "client_address=10.10.10.10",
        "client_name=mail.chapps.io",
        "reverse_client_name=mail.chapps.io",
        "instance=a483.61706bf9.17663.0",
        "sasl_method=plain",
        "sasl_username=somebody@chapps.io",
        "sasl_sender=",
        "size=12345",
        "ccert_subject=",
        "ccert_issuer=Caleb+20Cullen",
        "ccert_fingerprint=DE:AD:BE:EF:FE:ED:AD:DE:D0:A7:52:F3:C1:DA:6E:04",
        "encryption_protocol=TLSv1/SSLv3",
        "encryption_cipher=DHE-RSA-AES256-SHA",
        "encryption_keysize=256",
        "etrn_domain=",
        "stress=",
        "ccert_pubkey_fingerprint=68:B3:29:DA:98:93:E3:40:99:C7:D8:AD:5C:B9:C9:40",
        "client_port=1234",
        "policy_context=submission",
        "server_address=10.3.2.1",
        "server_port=54321",
        "",
      ]

    Refer to the `Postfix policy delegation
    documentation <http://www.postfix.org/SMTPD_POLICY_README.html>`
    for more information.

    As an example of the class's utility, and using the above definition of
    `payload`, consider:

    .. code:: python

      from chapps.util import PostfixPolicyRequest

      ppr = PostfixPolicyRequest(payload)

      # all the following are true:
      ppr.sender == 'unauth@easydns.com'
      ppr.sasl_username == 'somebody@chapps.io'
      ppr.client_address == '10.10.10.10'

      # demonstrating the pseudo-attribute:
      ppr.recipients == ['bar@foo.tld']

    Instance attributes (apart from Postfix payload parameters):

      :_payload: the Postfix policy delegation request payload in string-per-line format

      :recipients: a pseudo-attribute of the policy request derived from the
         value of `recipient`, provided by Postfix, which may contain more
         than one comma-separated email address.  For reasons unknown, Postfix
         always provides a `recipient_count` of 0 before the DATA phase, so
         we rely upon counting the email addresses directly.

      :_recipients: memoization attribute for :meth:`.recipients`

    .. admonition:: Subclassing

      Because the purpose of the class is to present the contents of the
      initial payload as attributes, all internal attributes are
      prefaced with an underscore.

    .. document private functions
    .. automethod:: __getattr__

    """

    def __init__(self, payload: List[str], *args, **kwargs):
        """Store the payload.

        :param List[str] payload: strings which are formatted as 'key=val',
          including an empty entry at the end.

        This routine discards the last element of the list and stores the rest
        as `self._payload`.

        """
        self._payload = payload[0:-2]

    # the main reason for this class:
    # find and memoize as attributes the values in the request payload
    # means we cannot test existence of attrs by getattr or self.<attr>
    def __getattr__(self, attr: str) -> Optional[str]:
        """Overloaded in order to search for missing attributes in the payload

        :param str attr: the attribute which triggered this call

        :returns: the value found in the payload, or :obj:`None`

        :rtype: Optional[str]

        First, if the value of `attr` starts with an underscore, `None` is
        returned.  No lines of the payload start with an underscore.  This
        ensures that references to internal attributes of the class are not
        snarled up with the payload searches.

        Next, the payload is searched for the requested key-value pair,
        attempting to match `attr` against everything before the `=` sign.
        When a line is found, the contents after the `=` are stored as an
        attribute named `attr` (and so memoized), and the value is returned.
        Future attempts to obtain the value will encounter the attribute and
        not invoke :meth:`.__getattr__` again.

        A `DEBUG` level message is currently produced if no lines in the
        payload matched the requested payload data.  No errors are produced
        if a nonexistent `attr` starting with `_` is encountered.

        """
        if attr[0] == "_":  # leading underscores do not occur in the payload
            return None
        line = next(
            (l for l in self._payload if attr == l.split("=")[0]), None
        )
        if line:
            key, *values = line.split("=")
            value = "=".join(values)
            setattr(self, key, value)
            return value
        else:
            logger.debug(f"No lines in {self} matched {attr}.")
        return None

    # Since the datastructure can function as a hash, provide optimization
    def __getitem__(self, key) -> Optional[str]:
        """
        Getting an item should optimize the result
        via the attribute mechanism
        """
        return getattr(self, key)

    # The datastructure is iterable, in case we need to enumerate the request
    def __iter__(self):
        """Return an iterable representing the mapping

        There should be few reasons to ever do this, though it comes in quite
        handy for testing.  This routine memoizes the dict it creates and also
        stores all the keys as attributes for future accesses.

        """
        if not getattr(self, "_mapping", None):
            self._mapping = {
                k: "=".join(vs)
                for k, *vs in [l.split("=") for l in self._payload]
            }
            # Since we end up parsing the entire payload
            # optimize it for future random access
            for k, v in self._mapping.items():
                setattr(self, k, v)
        yield from self._mapping

    # The length of the PPR is considered to be the number of items stored
    def __len__(self) -> int:
        """Act like a dict and return the number of k,v pairs

        :returns: number of lines in the payload

        :rtype: int

        """
        return len(self._payload)

    # Representations of PPR use their own class name, and otherwise dump
    # the payload, but not a list of attributes
    def __repr__(self) -> str:
        """Dump the data in _payload"""
        return "%s( %r )" % (self.__class__.__name__, self._payload)

    # In order to use memoization with PPRs, a hash function is required
    def __hash__(self):
        """Create a reliable hash for this PPR"""
        if "_hash" not in vars(self):
            self._hash = hash(f"{self.instance}:{self.queue_id}")
        return self._hash

    # This convenience property provides a list of recipient email addresses
    # since the line may contain more than one
    @property
    def recipients(self) -> List[str]:
        """Memoize recipients as a list

        :returns: a list of strings which are the email addresses of recipients

        :rtype: List[str]

        A convenience method to split the 'recipient' datum into
        comma-separated tokens for easier counting.

        """
        if "_recipients" not in vars(self):
            self._recipients = (
                self.recipient.split(",")
                if self.recipient and len(self.recipient) > 0
                else []
            )
        return self._recipients

    def domain_from(self, email_address: str) -> str:
        """Given an email address, return the domain part

        Raises meaningful errors if nonconforming conditions are encountered.
        """
        parts = email_address.split("@")
        if len(parts) > 2:
            logger.info(
                "Found sender email with more than one at-sign: "
                f"sender={email_address} instance={self.instance} "
                f"parts={parts!r}"
            )
            raise TooManyAtsException(f"{email_address}=>{parts!r}")
        elif len(parts) == 1:
            logger.info(
                "Found sender string without at-sign: "
                f"sender={email_address} instance={self.instance} "
                f"parts={parts!r}"
            )
            raise NotAnEmailAddressException
        return parts[-1]

    def helo_match(self, candidates: Dict[str, str]) -> bool:
        """HELO Whitelisting indicator

        Given a mapping of HELO name to expected IP address,
        indicate whether this PPR represents traffic from a
        whitelisted server.

        """
        name = self.helo_name or self.client_name
        if not (candidates and name in candidates):
            return False
        if candidates[name] == self.client_address:
            return True
        return False



"""
Communication handlers
----------------------

This module encapsulates the particular logic of:

  1. receiving data payloads from Postfix, and then

  2. sending back an appropriately-formatted response once the policy manager
     has had a chance to weigh in on the payload contents

Classes defined here exist mainly to be factories which return the main-loop
closure for :mod:`asyncio`.

TODO: :class:`.RequestHandler` should be a subclass of
`.CascadingPolicyHandler` which simply only ever has one policy within it.
This is to avoid maintaining two nearly-identical code-paths.  Running only one
policy is obviously a special case of running many.

"""
# from chapps.config import config  # the global instance of the config object
from chapps.policy import (
    EmailPolicy,
    OutboundQuotaPolicy,
    GreylistingPolicy,
    SenderDomainAuthPolicy,
)
from chapps.util import PostfixPolicyRequest
from chapps.inbound import InboundPPR
from chapps.outbound import OutboundPPR
from chapps.signals import (
    CHAPPSException,
    CallableExhausted,
    NullSenderException,
    AuthenticationFailureException,
)
from functools import cached_property
from typing import Type, Optional, List
import logging
import asyncio

try:
    from chapps.spf_policy import SPFEnforcementPolicy
except Exception:
    HAVE_SPF = False
    pass
else:
    HAVE_SPF = True

logger = logging.getLogger(__name__)  # pragma: no cover


class CascadingPolicyHandler:
    """Second-generation handler class which cascades multiple yes/no policies

    This class started out nearly identical to :class:`.RequestHandler`, but as
    the software has moved on, so has this handler, which is the main one in
    general use.

    This handler accepts a list of policy manager instances, all of which
    should produce True/False results.  The handler applies each policy to each
    request, and passes those which pass both, or returns the message from the
    policy which failed.  Once a policy has failed, no further policies are
    consulted.

    Instance attributes:

      :policies: a list of :class:`~chapps.policy.EmailPolicy` objects

      :pprclass: the class of :class:`~PostfixPolicyRequest` to instantiate
        from the Postfix request payload

      :config: a reference to the :class:`~chapps.config.CHAPPSConfig` stored
        on the first policy in the list

      :listen_address: the IP address to bind to; see :meth:`.listen_address`

      :listen_port: the port to listen on; see :meth:`.listen_port`

      :listener_backlog: how many connections may be pending (default: 100)

    """

    def __init__(
        self,
        policies: Optional[List[EmailPolicy]] = [],
        *,
        pprclass: PostfixPolicyRequest = PostfixPolicyRequest,
    ):
        self.policies = policies
        self.pprclass = pprclass
        if not self.policies:
            raise ValueError("A list of policy objects must be provided.")
        self.config = self.policies[
            0
        ].config  # all copies of the config are the same
        logger.debug(
            "Grabbing config from file "
            + self.config.chapps.config_file
            + " via "
            + self.policies[0].__class__.__name__
        )

    @cached_property
    def listen_address(self):
        return next(
            (getattr(p.params, "listen_address", None) for p in self.policies),
            None,
        )

    @cached_property
    def listen_port(self):
        return next(
            (getattr(p.params, "listen_port", None) for p in self.policies),
            None,
        )

    @cached_property
    def listener_backlog(self):
        return next(
            (
                getattr(p.config.chapps, "listener_backlog", None)
                for p in self.policies
            ),
            None,
        )

    # an asynchronous policy handler which cascades through all the policies;
    # fails stop execution
    def async_policy_handler(self):
        """Coroutine factory

        :returns: a coroutine which handles requests according to the policies,
          in order

        """
        pprclass = self.pprclass
        policies = self.policies
        encoding = self.config.chapps.payload_encoding
        no_user_key_response = self.config.chapps.no_user_key_response
        logger.debug(
            f"Cascading policy handler requested for "
            f"{[ type(p) for p in policies ]} using PPR "
            f"class {pprclass.__name__}."
        )

        async def handle_policy_request(reader, writer):
            """Handles reading and writing the streams around policy approval messages, and manages the cascade"""
            while True:
                try:
                    policy_payload = await reader.readuntil(b"\n\n")
                except ConnectionError:
                    logger.debug(
                        "Postfix said goodbye. Terminating this task."
                    )
                    return
                except asyncio.IncompleteReadError as e:
                    logger.debug(
                        "Postfix hung up before a read could be completed."
                        f"\n  Got: {e.partial}\n"
                        "Terminating this task."
                    )
                    return
                except CallableExhausted as e:
                    raise e
                except Exception:
                    if reader.at_eof():
                        logger.debug(
                            "Postfix said goodbye oddly. "
                            + "Terminating this task."
                        )
                        return
                    else:
                        logger.exception("UNEXPECTED ")
                    return
                logger.debug(
                    f"Payload received: {policy_payload.decode( 'utf-8' )}"
                )
                policy_data = pprclass(
                    policy_payload.decode(encoding).split("\n")
                )
                approval = True
                for policy in policies:
                    try:
                        if policy.approve_policy_request(policy_data):
                            resp = (
                                "action="
                                + policy.params.acceptance_message
                                + "\n\n"
                            )
                            logger.info(
                                f"{type(policy).__name__} PASS {policy_data}"
                            )
                        else:
                            resp = (
                                "action="
                                + policy.params.rejection_message
                                + "\n\n"
                            )
                            approval = False
                            logger.info(
                                f"{type(policy).__name__} FAIL {policy_data}"
                            )
                    except NullSenderException:
                        if policy.params.null_sender_ok:
                            resp = (
                                "action="
                                + policy.params.acceptance_message
                                + "\n\n"
                            )
                            logger.info(
                                f"{type(policy).__name__} PASS NS {policy_data}"
                            )
                        else:
                            resp = (
                                "action="
                                + policy.params.rejection_message
                                + "\n\n"
                            )
                            approval = False
                            logger.info(
                                f"{type(policy).__name__} FAIL NS {policy_data}"
                            )
                    except AuthenticationFailureException:
                        resp = "action=" + no_user_key_response + "\n\n"
                        approval = False
                        logger.info(
                            f"{type(policy).__name__} FAIL NA {policy_data}"
                        )
                    except CHAPPSException:
                        logger.exception("During policy evaluation:")
                    if not approval:
                        break
                try:
                    writer.write(resp.encode())
                except asyncio.CancelledError:
                    pass
                except Exception:
                    logger.exception(
                        f"Exception raised trying to send {resp.strip()}"
                    )
                    return

        return handle_policy_request


class RequestHandler(CascadingPolicyHandler):
    """Refactored intermediate base class for wrapping policy managers in an event loop

    This class has been reimplemented as a subclass of
    :class:`.CascadingPolicyHandler`, as a special case which has only one
    policy to handle.

    Instance attributes:

      :policy: the :class:`~chapps.policy.EmailPolicy` manager instance

      :config: a reference to the :class:`~chapps.config.CHAPPSConfig` stored
        on the policy instance.

      :pprclass: a reference to the particular kind of
        :class:`~chapps.util.PostfixPolicyRequest` to instantiate

    """

    def __init__(
        self,
        policy: EmailPolicy,
        *,
        pprclass: Type[PostfixPolicyRequest] = PostfixPolicyRequest,
    ):
        """Setup a Postfix policy request handler

        :param chapps.policy.EmailPolicy policy: an instance of a policy
          manager (a subclass of :class:`~chapps.policy.EmailPolicy`)

        :param Type[PostfixPolicyRequest] pprclass: the subclass of :class:`~chapps.util.PostfixPolicyRequest` to instantiate from the Postfix payloads; defaults to :class:`~chapps.util.PostfixPolicyRequest`

        .. note::

          Unlike other class families within CHAPPS, the handlers in this
          module do not accept config-override arguments.  They obtain their
          references to the config from their attached policy managers.

        """
        super().__init__([policy], pprclass=pprclass)

    @cached_property
    def listen_address(self):
        return self.policy.params.listen_address

    @cached_property
    def listen_port(self):
        return self.policy.params.listen_port

    @cached_property
    def policy(self):
        return self.policies[0]


class OutboundMultipolicyHandler(CascadingPolicyHandler):
    """Convenience subclass for combining outbound P/F policies

    Could be thought of as a concrete subclass of
    :class:`~.CascadingPolicyHandler`, but meant more as a convenience.

    """

    def __init__(self, policies=[], *, pprclass=OutboundPPR):
        """Setup an OutboundMultipolicyHandler

        :param List[EmailPolicy] policies: a list of policy manager instances

        :param Type[PostfixPolicyRequest] pprclass: kind of
          :class:`~chapps.util.PostfixPolicyRequest` to instantiate from
          Postfix request payloads; defaults to
          :class:`~chapps.outbound.OutboundPPR`

        If none are provided, default-configured instances of
        :class:`~chapps.policy.SenderDomainAuthPolicy` and
        :class:`~chapps.policy.OutboundQuotaPolicy` are used, in that order.

        """
        # note that we default to OutboundPPR here
        policies = policies or [
            SenderDomainAuthPolicy(),
            OutboundQuotaPolicy(),
        ]  # create a list of relevant outbound y/n policies
        super().__init__(policies, pprclass=pprclass)


class OutboundQuotaHandler(RequestHandler):
    """Convenience class for wrapping :class:`~chapps.policy.OutboundQuotaPolicy`"""

    def __init__(self, policy: OutboundQuotaPolicy = None):
        """Setup an OutboundQuotaHandler

        :param chapps.policy.OutboundQuotaPolicy policy: an instance of
          :class:`~chapps.policy.OutboundQuotaPolicy`

        """
        p = policy or OutboundQuotaPolicy()
        super().__init__(p, pprclass=OutboundPPR)


class GreylistingHandler(RequestHandler):
    """Convenience class for wrapping :class:`~chapps.policy.GreylistingPolicy`"""

    def __init__(self, policy: GreylistingPolicy = None):
        """Setup a GreylistingHandler

        :param chapps.policy.GreylistingPolicy policy: an instance of
          :class:`~chapps.policy.GreylistingPolicy`

        """
        p = policy or GreylistingPolicy()
        super().__init__(p, pprclass=InboundPPR)


class SenderDomainAuthHandler(RequestHandler):
    """Convenience class for wrapping :class:`~chapps.policy.SenderDomainAuthPolicy`"""

    def __init__(self, policy: SenderDomainAuthPolicy = None):
        """Setup a SenderDomainAuthHandler

        :param chapps.policy.SenderDomainAuthPolicy policy: an instance of
          :class:`~chapps.policy.SenderDomainAuthPolicy`

        """
        p = policy or SenderDomainAuthPolicy()
        super().__init__(p, pprclass=OutboundPPR)


class CascadingMultiresultPolicyHandler(CascadingPolicyHandler):
    """Cascading handler for policies which produce more than two results

    Some policies, such as the SPF enforcement policy, need to be able to
    generate responses with more nuance than simply pass/fail, and so there is
    a need for a handler which can deal with policies which return strings
    (Postfix directives), or possibly a custom class to encapsulate also some
    idea of pass/fail, in order to know whether to abort the policy-evaluation
    loop early.

    """

    def async_policy_handler(self):
        """Returns a coroutine which handles results according to the configuration

        The policy being enforced is stored in the SPF-related TXT record
        on the sender's domain.  The local configuration of this policy
        amounts to instructions about responses to different outcomes of
        the SPF check, along with what IP address and port to listen on.

        This policy handler is different from others in that, because it
        does not expect a PASS/FAIL response, it simply wraps the return
        value of
        :func:`~chapps.spf_policy.SPFEnforcementPolicy.approve_policy_request()`
        in a Postfix response packet, and sends it.  Rather than refer to
        pre-configured acceptance and rejection messages, it expects the
        approval routine to send a string which can be interpreted by
        Postfix as a command.

        TODO: In order to be able to cascade *through* this kind of policy,
        it is going to have to return a first-class object which can be
        annotated as being a pass or a fail, so that a cascading handler
        can decide whether to continue.  That object's `__str__()`
        method will need to return the Postfix command.

        .. todo::

          In order to support more than one MTA (tho no such support is
          planned), the action-translation layer might be refactored out of
          the policy itself, to be applied here, in order to switch between
          different types.

        """
        ### This override version for SPF enforcement does not assume a yes-or-no response pattern
        logger.debug(
            "Policy handler requested for"
            f" {[type(p).__name__ for p in self.policies]!r}."
        )
        pprclass = self.pprclass
        policies = self.policies
        encoding = self.config.chapps.payload_encoding

        logger.debug(
            "Cascading Multipolicy handler requested for "
            f"{[type(p) for p in policies]} using PPR "
            f"class {pprclass.__name__}."
        )

        async def handle_policy_request(reader, writer):
            """Handles reading and writing the streams around policy approval messages"""

            async def close_streams():
                writer.close()
                await writer.wait_closed()

            while True:
                try:
                    policy_payload = await reader.readuntil(b"\n\n")
                except ConnectionError:
                    logger.debug(
                        "Postfix said goodbye. Terminating this task."
                    )
                    close_streams()
                    return
                except asyncio.IncompleteReadError as e:
                    logger.debug(
                        "Postfix hung up before a read could be completed."
                        f"\n  Got: {e.partial}\n"
                        "Terminating this task."
                    )
                    close_streams()
                    return
                except (CallableExhausted, InterruptedError) as e:
                    raise e
                except Exception as e:
                    if reader.at_eof():
                        logger.debug(
                            "Postfix said goodbye oddly: {e}"
                            " Terminating this task."
                        )
                    else:
                        logger.exception("UNEXPECTED ")
                    close_streams()
                    return  # prev: continue
                logger.debug(
                    f"Payload received: {policy_payload.decode(encoding)}"
                )
                policy_data = pprclass(
                    policy_payload.decode(encoding).split("\n")
                )

                actions = ["DUNNO"]
                # TODO:
                # track all responses, and then extract non-DUNNO ones if any
                for policy in policies:
                    try:
                        action = policy.approve_policy_request(policy_data)
                        actions.append(action)
                        logger.info(
                            f"{type(policy).__name__} "
                            # + ("PASS" if action else "FAIL")
                            + f" {action!r} {policy_data}"
                        )
                    except CHAPPSException:
                        logger.exception("During policy evaluation:")
                    if not action:
                        break

                if action:  # i.e. the mail will be forwarded
                    non_dunno = [a for a in actions if a != "DUNNO"]
                    action = non_dunno[-1] if non_dunno else "DUNNO"
                try:
                    writer.write((f"action={action}\n\n").encode(encoding))
                except asyncio.CancelledError:  # pragma: no cover
                    pass
                except Exception:
                    logger.exception(
                        f"Exception raised trying to send {action}"
                    )
                    close_streams()
                    return
                else:
                    await writer.drain()

        return handle_policy_request


class MultiresultPolicyHandler(CascadingMultiresultPolicyHandler):
    """A subclass for handling just one policy"""

    def __init__(
        self,
        policy: EmailPolicy,
        *,
        pprclass: Optional[PostfixPolicyRequest] = PostfixPolicyRequest,
    ):
        super().__init__([policy], pprclass=pprclass)


if HAVE_SPF:

    class SPFEnforcementHandler(MultiresultPolicyHandler):
        """Special handler class for :class:`~chapps.spf_policy.SPFEnforcementPolicy`

        This one came along last and forced a reconsideration of how all this
        worked, because it produces more than two possible states as output.
        The plan is to retrofit all the older policies so that they also can
        use an action-translation layer, but that will also require some
        adjustment of the cascading handler.

        .. note::

          This class will not be defined if the relevant SPF libraries could
          not be loaded.  They may be installed via `pip` using the extras
          mechanism: ``pip install chapps[SPF]``

        """

        def __init__(
            self,
            policy: Optional[SPFEnforcementPolicy] = None,
            pprclass: Optional[PostfixPolicyRequest] = InboundPPR,
        ):
            """Set up an SPFEnforcementHandler

            :param chapps.spf_policy.SPFEnforcementPolicy policy: an instance
              of :class:`~chapps.spf_policy.SPFEnforcementPolicy`

            """
            p = policy or SPFEnforcementPolicy()
            super().__init__(p, pprclass=pprclass)

        @cached_property
        def policy(self):
            return self.policies[0]

    class InboundMultipolicyHandler(CascadingMultiresultPolicyHandler):
        """Implements SPF and Greylisting simultaneously

        This is a template for an inbound multipolicy handler, which by default
        checks SPF and also performs Greylisting, each of which depends upon
        options which the email account administrator may set about whether
        either of SPF or Greylisting or both should be applied to inbound mail
        for the particular domain.

        """

        def __init__(
            self,
            policies: Optional[List[EmailPolicy]] = None,
            *,
            pprclass: Optional[PostfixPolicyRequest] = InboundPPR,
        ):
            """Create an inbound policy handler for SPF + Greylisting"""
            policies = policies or [
                SPFEnforcementPolicy(),
                GreylistingPolicy(),
            ]
            super().__init__(policies, pprclass=pprclass)
